from numpy import ones, array, sqrt, trapz
from scipy.sparse import dia_matrix
from scipy.sparse.linalg import eigs
from solcore.science_tracker import science_reference
from solcore.constants import *
import numpy as np

from operator import itemgetter

sort_simultaneous = lambda *lists: [list(x) for x in zip(*sorted(zip(*lists), key=itemgetter(0)))]

from . import structure_utilities
from .graphics import Graph, GraphData


def tridiag_euler(V, z, m, periodic=False, num_eigenvalues=10, quasiconfined=0):
    """
    Returns eignvalue and eigenvectors of an arbitrary potential.
    
    A tridiagonal matrix is constructed by writing the variable effective
    mass Schrodinger equation over a series of mesh points. The eigenvalues
    of the matrix correspond to the allowed energy levels of the system.
    
    The previous solver, eig, has been replaced by the spare matrix version, eigs, that is faster to compute
    """

    science_reference("Varible effective mass Schordinger equation and tridiagonal solution method.",
                      "Frensley, W. R. (1991). \
                      Numerical evaluation of resonant states. \
                      Superlattices and Microstructures, 11(3), 347350. \
                      doi:10.1016/0749-6036(92)90396-M")

    N = len(V)
    dz = np.gradient(z)
    m = m * ones(N)

    # Vectorise effective mass differences to avoid a loop
    m = np.insert(m, (0, len(m)), (m[0], m[-1]))
    m_a = m[0:-2]  # m_(j-1) from Frensley, W. R. (1991)
    m_b = m[1:-1]  # m_j
    m_c = m[2:]  # m_(j+1)

    # These are the interior diagonals of equation 18 in the above ref.
    axis = hbar ** 2 / (4 * dz ** 2) * (1 / m_a + 2 / m_b + 1 / m_c) + V  # d_j from Frensley, W. R. (1991)
    upper = hbar ** 2 / (4 * dz ** 2) * (1 / m_a + 1 / m_b)  # s_(j+1)
    lower = hbar ** 2 / (4 * dz ** 2) * (1 / m_b + 1 / m_c)  # s_j

    if periodic:
        TopRight = np.zeros(len(axis))
        BottomLeft = np.zeros(len(axis))
        TopRight[-1] = -lower[-1]  # The last point becomes the one before the first one
        BottomLeft[0] = -upper[0]  # The first point becomes the one after the last one

        index = (-N + 1, -1, 0, 1, N - 1)
        diagonals = (BottomLeft, -lower, axis, -upper, TopRight)

    else:
        index = (-1, 0, 1)
        diagonals = (-lower, axis, -upper)

    H = dia_matrix((diagonals, index), shape=(N, N))

    # H[0,-1] = -lower[-1]    # The last point becomes the one before the first one
    # H[-1,0] = -upper[0]     # The first point becomes the one after the last one

    sigma = np.min(V)  # The top of the potential (valence band) is the target energy for the eigenvalues

    # The heavy numerical calculation
    E, Psi = eigs(H, k=num_eigenvalues, which='LR', sigma=sigma)

    # Allow for quasi confined levels to go through. They can be discarded later with the filter
    confined_levels = [i for i, e in enumerate(E) if e < quasiconfined * q]
    E, Psi = E[confined_levels].real, array(Psi[:, confined_levels]).transpose()
    Psi = [(p / sqrt(trapz(p * p, x=z))).real for p in Psi]

    E, Psi = sort_simultaneous(E, Psi)

    return E, Psi


def schroedinger_solve(x, V, m, num_eigenvalues=10, periodic=False, offset=0, electron=True, quasiconfined=0):
    """Returns normalised wavefuctions from the potential profile.

    Arguments:

        x -- spatial grid
        V -- potential
        m -- effective mass

    Keywords:

        electron -- whether the wavefunctions describe electrons or holes (default: True)
        num_eigenvalues -- Number of eigenvalues to calculate (default = 10)
        periodic   -- not to sure what this does (default: False)
    """

    if electron:
        v_shift = max(V) + offset
        E, psi = tridiag_euler(V - v_shift, x, m, num_eigenvalues=num_eigenvalues, periodic=periodic,
                               quasiconfined=quasiconfined)
        E = list() if len(E) == 0 else np.array(E) + v_shift

    else:
        v_shift = min(V) - offset
        E, psi = tridiag_euler(-V + v_shift, x, m, num_eigenvalues=num_eigenvalues, periodic=periodic,
                               quasiconfined=quasiconfined)
        E = list() if len(E) == 0 else -np.array(E) + v_shift

    return E, psi


def __potentials_to_wavefunctions_energies_internal(x, Ve, me, Vhh, mhh, Vlh, mlh, num_eigenvalues=10, periodic=False,
                                                    offset=0, filter_strength=0, structure=None, quasiconfined=0):
    """Returns normalised wavefuctions from the potential profile.

    Arguments:

        x -- spatial grid
        Ve -- electron potential
        me -- electron effective mass
        Vlh -- light hole potential
        mlh -- light hole effective mass
        Vhh -- heavy hole potential
        mhh -- heavy hole effective mass


    Keywords:

        num_eigenvalues -- Number of eigenvalues to calculate (default = 10)
        periodic   -- not to sure what this does (default: False)
    """

    Ee, psi_e = schroedinger_solve(x, Ve, me, num_eigenvalues, periodic, offset, electron=True,
                                   quasiconfined=quasiconfined)
    Ehh, psi_hh = schroedinger_solve(x, Vhh, mhh, num_eigenvalues, periodic, offset, electron=False,
                                     quasiconfined=quasiconfined)
    Elh, psi_lh = schroedinger_solve(x, Vlh, mlh, num_eigenvalues, periodic, offset, electron=False,
                                     quasiconfined=quasiconfined)

    if filter_strength != 0:
        assert structure is not None, "Need to provide structure to find well regions for filtering"
        print("filtering")
        Ee, psi_e = discard_unconfined(x, structure, Ee, psi_e, filter_strength)
        Ehh, psi_hh = discard_unconfined(x, structure, Ehh, psi_hh, filter_strength)
        Elh, psi_lh = discard_unconfined(x, structure, Elh, psi_lh, filter_strength)

    # Ee, psi_e = discard_unconfined_energy(Ee, psi_e, Ve, quasiconfined)
    # Ehh, psi_hh = discard_unconfined_energy(Ehh, psi_hh, Vhh, quasiconfined)
    # Elh, psi_lh = discard_unconfined_energy(Elh, psi_lh, Vlh, quasiconfined)

    # print(me/electron_mass, mhh/electron_mass, mlh/electron_mass)
    # print(me_plane, mhh_plane, mlh_plane)
    #
    # import sys
    # sys.exit()

    return x, Ee, psi_e, Ehh, psi_hh, Elh, psi_lh


def discard_unconfined_energy(E, psi, V, quasiconfined):
    before = len(E)
    maxE = min(abs(V[0]), max(abs(V))) + quasiconfined
    try:
        E, psi = zip(*[(E_i, psi_i)
                       for (E_i, psi_i) in zip(E, psi)
                       if abs(E_i) <= maxE])
    except ValueError as exception:
        print("Warning: wavefunction filter removed all states for this band, try reducing the filter strength.")
        return ([], [])
    print("Wavefunction filter removed %d state%s." % (before - len(E), "" if (before - len(E)) == 1 else "s"))
    return E, psi


def discard_unconfined(x, structure, E, psi, threshold=0.8):
    if threshold == 0:  # bypass the filter code, saving time
        return E, psi

    indx = structure_utilities.well_regions(x, structure)

    before = len(E)
    # for (E_i, psi_i) in zip(E, psi): 
    #     print (psi_i[indx])

    try:

        E, psi = zip(*[(E_i, psi_i)
                       for (E_i, psi_i) in zip(E, psi)
                       if np.trapz(psi_i[indx] ** 2, x=x[indx]) / np.trapz(psi_i ** 2, x=x) >= threshold])
    except ValueError as exception:
        print("Warning: wavefunction filter removed all states for this band, try reducing the filter strength.")
        return ([], [])
    print("Wavefunction filter removed %d state%s." % (before - len(E), "" if (before - len(E)) == 1 else "s"))
    return E, psi


def potentials_to_wavefunctions_energies(x, Ve, me, Vhh, mhh, Vlh, mlh, num_eigenvalues=10, periodic=False, offset=0,
                                         filter_strength=0, structure=None, quasiconfined=0, **kwargs):
    x, Ee, psi_e, Ehh, psi_hh, Elh, psi_lh = __potentials_to_wavefunctions_energies_internal(
        x, Ve, me, Vhh, mhh, Vlh, mlh, num_eigenvalues,
        periodic, offset,
        filter_strength, structure,
        quasiconfined)
    return {
        "x": x,
        "Ee": Ee,
        "psi_e": psi_e,
        "Ehh": Ehh,
        "psi_hh": psi_hh,
        "Elh": Elh,
        "psi_lh": psi_lh,
        "Ve": Ve,
        "me": me,
        "Vhh": Vhh,
        "Vlh": Vlh,
        "mhh": mhh,
        "mlh": mlh
    }


def graph(x, Ve, me, Vhh, mhh, Vlh, mlh, **kwargs):
    defaults = {
        "edit": lambda x, y: (x * 1e9, y / q),
        "xlabel": "Depth (nm)",
        "ylabel": "Energy (eV)",
    }
    defaults.update(kwargs)
    data = []

    normalise_psi = lambda p: p * q * 5e-6

    data.append(GraphData(x, Ve, color="black", linewidth=2))
    data.append(GraphData(x, Vlh, color="black", linewidth=2, dashes=[1, 1]))
    data.append(GraphData(x, Vhh, color="black", linewidth=2))

    g = Graph(data, **defaults)
    return g



from solcore.quantum_mechanics.kp_QW import solve_bandstructure_QW
from solcore.quantum_mechanics.heterostructure_alignment import VBO_align
from solcore.quantum_mechanics.structure_utilities import structure_to_potentials
from solcore.quantum_mechanics.potential_utilities import potentials_to_wavefunctions_energies
from solcore.quantum_mechanics import graphics
from solcore.absorption_calculator.absorption_QW import calc_alpha
from solcore.interpolate import interp1d
from solcore.constants import q, vacuum_permittivity
import numpy as np


def schrodinger(structure, plot_bands=False, kpoints=40, krange=1e9, num_eigenvalues=10, symmetric=True,
                quasiconfined=0.0, return_qw_boolean_for_layer=False, Efield=0, blur=False, blurmode="even",
                mode='kp8x8_bulk', step_size=None, minimum_step_size=0, smallest_feature_steps=20, filter_strength=0,
                periodic=False, offset=0, graphtype=[], calculate_absorption=False, alpha_params=None, **kwargs):
    """ Solves the Schrodinger equation of a 1 dimensional structure. Depending on the inputs, the method for solving
    the problem is more or less sophisticated. In all cases, the output includes the band structure and effective
    masses around k=0 as a function of the position, the energy levels of the QW (electrons and holes) and the
    wavefunctions, although for the kp4x4 and kp6x6 modes, this is provided as a function of the k value, and therefore
    there is much more information.

    :param structure: The strucutre to solve
    :param plot_bands: (False) If the bands should be plotted
    :param kpoints: (30) The number of points in the k space
    :param krange: (1e-9) The range in the k space
    :param num_eigenvalues: (10) Maximum number of eigenvalues to calculate
    :param symmetric: (True) If the structure is symmetric, in which case the calculation can be speed up
    :param quasiconfined: (0.0 eV) Energy above the band edges that an energy level can have before rejecting it
    :param return_qw_boolean_for_layer: (False) Return an boolean array indicating which positions are inside the QW
    :param Efield: (0) Electric field.
    :param blur: (False) If the potentials and effective masses have to be blurred
    :param blurmode: ('even') Other values are 'right' and 'left'
    :param mode: ('kp4x4') The mode of calculating the bands and effective masses. See 'structure_utilities.structure_to_potentials'
    :param step_size: (None) The discretization step of the structure. If none, it is estimated based on the smallest feature.
    :param minimum_step_size: (0) The minimum step size.
    :param smallest_feature_steps: (20) The number of steps in the smallest feature
    :param filter_strength: (0) If > 0, defines the fraction of the wavefunction that has to be inside the QW in order to consider it 'confined'
    :param periodic: (False) If the structure is periodic. Affects the boundary conditions.
    :param offset: (0) Energy offset used in the calculation of the energy levels in the case of the 'bulk' solvers
    :param graphtype: [] If 'potential', the band profile and wavefunctions are ploted
    :return: A dictionary containing the band structure and wavefunctions as a function of the position and k
    """

    if Efield != 0:
        symmetric = False

    aligned_structure = VBO_align(structure)
    potentials = structure_to_potentials(aligned_structure, return_qw_boolean_for_layer=return_qw_boolean_for_layer,
                                         Efield=Efield, blur=blur, blurmode=blurmode, mode=mode, step_size=step_size,
                                         minimum_step_size=minimum_step_size,
                                         smallest_feature_steps=smallest_feature_steps)

    # Now that we have the potential and effective masses ready, we solve the problem.
    if mode in ['kp4x4', 'kp6x6']:
        bands = solve_bandstructure_QW(potentials, num=num_eigenvalues, kpoints=kpoints, krange=krange,
                                       symmetric=symmetric, quasiconfined=quasiconfined, plot_bands=plot_bands)

        result_band_edge = {
            "x": bands['x'],
            "potentials": {key: potentials[key] for key in potentials.keys() if key[0] in "Vx"},
            "effective_masses": {key: potentials[key] for key in potentials.keys() if key[0] in "mx"},
            "wavefunctions": {key: bands[key][:, 0] for key in bands.keys() if 'psi' in key},
            "E": {key: bands[key][:, 0] for key in bands.keys() if key[0] in 'E'},
        }

    else:
        bands = potentials_to_wavefunctions_energies(structure=structure, num_eigenvalues=num_eigenvalues,
                                                     filter_strength=filter_strength, offset=offset, periodic=periodic,
                                                     quasiconfined=quasiconfined, **potentials)

        result_band_edge = {
            "x": bands['x'],
            "potentials": {key: potentials[key] for key in potentials.keys() if key[0] in "Vx"},
            "effective_masses": {key: potentials[key] for key in potentials.keys() if key[0] in "mx"},
            "wavefunctions": {key: bands[key] for key in bands.keys() if 'psi' in key},
            "E": {key: np.array(bands[key]) for key in bands.keys() if key[0] in "E"},
        }

    if "potentials" is graphtype:
        schrodinger_plt = graphics.split_schrodinger_graph_potentials(result_band_edge, **kwargs)
        # schrodinger_plt.draw()

    if "potentialsLDOS" is graphtype:
        Ee, LDOSe, Eh, LDOSh = graphics.split_schrodinger_graph_LDOS(result_band_edge, **kwargs)
        result_band_edge["LDOS"] = {"x": bands['x'], 'Ee': Ee, 'LDOSe': LDOSe, 'Eh': Eh, 'LDOSh': LDOSh}

    if calculate_absorption:
        result_band_edge["alpha"] = calc_alpha(result_band_edge, **alpha_params)
        result_band_edge["alphaE"] = interp1d(x=result_band_edge["alpha"][0], y=result_band_edge["alpha"][1])

    return result_band_edge, bands


if __name__ == "__main__":
    from solcore import si, material
    from solcore.structure import Layer, Structure
    import matplotlib.pyplot as plt
    import numpy as np

    bulk = material("GaAs")(T=293)
    barrier = material("GaAsP")(T=293, P=0.1)

    bulk.strained = False
    barrier.strained = True

    top_layer = Layer(width=si("30nm"), material=bulk)
    inter = Layer(width=si("3nm"), material=bulk)
    barrier_layer = Layer(width=si("15nm"), material=barrier)
    bottom_layer = top_layer

    E = np.linspace(1.15, 1.5, 300) * q
    alfas = np.zeros((len(E), 6))

    alfas[:, 0] = E / q

    alpha_params = {
        "well_width": si("7.2nm"),
        "theta": 0,
        "eps": 12.9 * vacuum_permittivity,
        "espace": E,
        "hwhm": si("6meV"),
        "dimensionality": 0.16,
        "line_shape": "Gauss"
    }

    comp = [0.05, 0.10, 0.15, 0.20]
    colors = plt.cm.jet(np.linspace(0, 1, len(comp)))

    plt.figure(figsize=(6, 4.5))
    for j, i in enumerate(comp):
        QW = material("InGaAs")(T=293, In=i)
        QW.strained = True
        well_layer = Layer(width=si("7.2nm"), material=QW)

        # test_structure = Structure([top_layer, barrier_layer, inter] + 1 * [well_layer, inter, barrier_layer, inter] +
        #                            [bottom_layer])

        test_structure = Structure([barrier_layer, inter] + 1 * [well_layer, inter] +
                                   [barrier_layer])

        # test_structure = Structure([top_layer, barrier_layer] + 10 * [well_layer, barrier_layer] +
        #                            [bottom_layer])

        test_structure.substrate = bulk

        output = schrodinger(test_structure, quasiconfined=0,  mode='kp4x4', plot_bands=False,
                             num_eigenvalues=20, alpha_params=alpha_params, calculate_absorption=True)

        alfa = output[0]['alphaE'](E)
        plt.plot(1240 / (E / q), alfa / 100, label='{}%'.format(int(i * 100)))
        alfas[:, j + 1] = alfa / 100

    plt.xlim(826, 1100)
    plt.ylim(0, 23000)
    plt.xlabel('Wavelength (nm)')
    plt.ylabel('$\\alpha$ cm$^{-1}$')
    plt.legend(loc='upper right', frameon=False)
    plt.tight_layout()

    import os

    root = os.path.expanduser('~')
    plt.savefig(root + '/Desktop/abs.pdf')
    plt.show()



from solcore.structure import Layer, Junction, TunnelJunction
from solcore.absorption_calculator import calculate_rat_rcwa, calculate_absorption_profile_rcwa

import numpy as np
import types
from solcore.state import State


rcwa_options = State()
rcwa_options.size = [500, 500]
rcwa_options.orders = 4
rcwa_options.theta = 0
rcwa_options.phi = 0
rcwa_options.pol = 'u'
rcwa_options.parallel = False
rcwa_options.n_jobs = -1

def solve_rcwa(solar_cell, options):
    """Calculates the reflection, transmission and absorption of a solar cell object using the rigorous coupled-wave analysis solver.

    :param solar_cell: A solar_cell object
    :param options: Options for the solver
    :return: None
    """
    wl = options.wavelength
    solar_cell.wavelength = options.wavelength
    parallel = options.parallel
    rcwa_options = options.get("rcwa_options", None)

    # We include the shadowing losses
    initial = (1 - solar_cell.shading) if hasattr(solar_cell, 'shading') else 1

    # Now we calculate the absorbed and transmitted light. We first get all the relevant parameters from the objects
    all_layers = []
    n_layers_junction = []
    for j, layer_object in enumerate(solar_cell):

        # Attenuation due to absorption in the AR coatings or any layer in the front that is not part of the junction
        if type(layer_object) is Layer:
            all_layers.append(layer_object)
            n_layers_junction.append(1)

        # For each junction, and layer within the junction, we get the absorption coefficient and the layer width.
        elif type(layer_object) in [TunnelJunction, Junction]:
            n_layers_junction.append(len(layer_object))
            for i, layer in enumerate(layer_object):
                all_layers.append(layer)

    # With all the information, we create the optical stack
    stack = all_layers

    position = options.position * 1e9

    substrate = solar_cell.substrate
    incidence = solar_cell.incidence

    print('Calculating RAT...')
    RAT = calculate_rat_rcwa(stack, options.size, options.orders, wl * 1e9, incidence, substrate, theta=options.theta,
                             phi=options.phi, pol=options.pol, parallel=parallel, user_options=rcwa_options)


    print('Calculating absorption profile...')
    out = calculate_absorption_profile_rcwa(stack, options.size, options.orders, wl*1e9, RAT['A_pol'],
                                      dist=position, theta=options.theta, phi=options.phi, pol=options.pol, incidence=incidence,
                                      substrate=substrate,
                                      parallel=options.parallel, n_jobs=options.n_jobs, user_options=rcwa_options)
    # With all this information, we are ready to calculate the differential absorption function
    diff_absorption, all_absorbed = calculate_absorption_rcwa(out, initial)


    layer = 0
    A_per_layer = np.array(RAT['A_per_layer'].T)

    # Each building block (layer or junction) needs to have access to the absorbed light in its region.
    # We update each object with that information.
    for j in range(len(solar_cell)):
        solar_cell[j].diff_absorption = diff_absorption
        solar_cell[j].absorbed = types.MethodType(absorbed, solar_cell[j])

        solar_cell[j].layer_absorption = initial*np.sum(A_per_layer[layer:(layer+n_layers_junction[j])], axis = 0)
        layer = layer + n_layers_junction[j]

    solar_cell.reflected = RAT['R'] * initial
    solar_cell.absorbed = sum([solar_cell[x].layer_absorption for x in np.arange(len(solar_cell))])
    solar_cell.transmitted = initial - solar_cell.reflected - solar_cell.absorbed


def absorbed(self, z):
    out = self.diff_absorption(self.offset + z) * (z < self.width)
    return out.T


def calculate_absorption_rcwa(tmm_out, initial=1):
    all_z = tmm_out['position'] * 1e-9
    all_abs = initial * tmm_out['absorption'] / 1e-9

    def diff_absorption(z):
        idx = all_z.searchsorted(z)
        idx = np.where(idx <= len(all_z) - 2, idx, len(all_z) - 2)
        try:
            z1 = all_z[idx]
            z2 = all_z[idx + 1]

            f = (z - z1) / (z2 - z1)

            out = f * all_abs[:, idx] + (1 - f) * all_abs[:, idx + 1]

        except IndexError:
            out = all_abs[:, idx]

        return out

    all_absorbed = np.trapz(diff_absorption(all_z), all_z)

    return diff_absorption, all_absorbed



from solcore.structure import Layer, Junction, Structure, TunnelJunction
from solcore import material, si

import numpy as np


def default_GaAs(T):
    # We create the other materials we need for the device
    window = material('AlGaAs')(T=T, Na=5e24, Al=0.8)
    p_GaAs = material('GaAs')(T=T, Na=1e24)
    n_GaAs = material('GaAs')(T=T, Nd=8e22)
    bsf = material('GaAs')(T=T, Nd=2e24)

    output = Junction([Layer(width=si('30nm'), material=window, role="Window"),
                       Layer(width=si('150nm'), material=p_GaAs, role="Emitter"),
                       Layer(width=si('3000nm'), material=n_GaAs, role="Base"),
                       Layer(width=si('200nm'), material=bsf, role="BSF")], sn=1e6, sp=1e6, T=T, kind='PDD')

    return output


class SolarCell(Structure):
    """ This class is almost identical to the basic Structure class in Solcore (it is a subclass of it, actually) but implementing some default parameter values and control about the types of layers. It should work anywhere where a Structure object works.
    """

    def __init__(self, layers=None, T=298, cell_area=1, reflectivity=None, shading=0, substrate=None, incidence=None,
                 R_series=0, **kwargs):
        """ Constructor of the class.

        :param layers: A list with the layers to add. The layers might include individual Layer objects of whole Junction objects.
        :param T: Temperature.
        :param cell_area: The area of the cell.
        :param reflectivity: Function that calculates the reflectivity as a function of energy.
        :param shading: Shading losses due to the front metal contacts.
        :param substrate: Substrate of the solar cell.
        :param incidence: Material above the solar cell. Nominally air.
        :param R_series: Series resistance of the structure
        :param kwargs: Other possible attributes.
        """

        assert isinstance(layers, list), "Layers must be provided inside a list, even if it is just one layer."

        super(SolarCell, self).__init__(layers)
        self.__dict__.update(kwargs)

        self.T = T
        self.cell_area = cell_area
        self.shading = shading
        self.reflectivity = reflectivity
        self.junctions = 0
        self.junction_indices = []
        self.tunnel_indices = []
        self.substrate = substrate
        self.incidence = incidence
        self.R_series = R_series

        for i, element in enumerate(layers):
            self.sort_layer_type(element, i)
            self.R_series += element.R_series if hasattr(element, 'R_series') else 0

    def sort_layer_type(self, layer, i):
        """ Sorts the layer in different categories, depending on its type, and keeps record on the indices of that type of layer.

        :param layer: The layer to check
        :param i: The index of that layer
        :return: None
        """

        if type(layer) == Junction:
            self.junction_indices.append(i)
            self.junctions += 1
        if type(layer) == TunnelJunction:
            self.tunnel_indices.append(i)

    def append(self, new_layer, layer_label=None, repeats=1):
        """ Appends a layer to the structure a certain number of times to the structure.

        :param new_layer: The layer to append.
        :param layer_label: An optional label for that layer.
        :param repeats: Number of times to repeat add the layer.
        :return: None
        """

        for i in range(repeats):
            self.sort_layer_type(new_layer, len(self))
            super(SolarCell, self).append(new_layer, layer_label)

    def append_multiple(self, layers, layer_labels=None, repeats=1):
        """ Appends multiple layers a certain umber of times to the structure.

        :param layers: A list with the layers to append.
        :param layer_labels: An optional list with the labels. If present, it must have the same length that the layers list.
        :param repeats: Number of times to add this set of layers.
        :return: None
        """

        assert isinstance(layers, list), "'append_multiple' only accepts lists for the first argument."

        if layer_labels is not None:
            assert len(layers) == len(layer_labels), "When using 'layer_labels' keyword a label must be specified for " \
                                                     "each layer added i.e. layers and layer_labels must have the same " \
                                                     "number of elements. Either fix this or simply do not assign any " \
                                                     "labels (i.e. layer_labels=None)."
        else:
            layer_labels = [None] * len(layers)

        for i in range(repeats):
            for j, element in enumerate(layers):
                self.append(element, layer_labels[j])

    def update_junction(self, junction, **kwargs):
        """ Adds or updates the attributes - not the layers - of a junction.

        :param junction: The junction to update.
        :param kwargs: The attributes to update.
        :return: None
        """

        try:
            num = self.junction_indices[junction]
            self[num].__dict__.update(kwargs)
        except IndexError:
            print('ERROR updating junction: The junction index must be {} or less.'.format(len(self.junction_indices)))

    def __call__(self, i):
        return self[self.junction_indices[i]]


if __name__ == '__main__':
    window = material('AlGaAs')(T=298, Na=1e24, Al=0.8)
    stack = [Layer(width=si("50nm"), material=window),
             default_GaAs(298)]

    # stack = [default_GaAs(298)]

    my_cell = SolarCell(layers=stack)
    #
    # my_cell.append_multiple([default_GaAs(298), default_GaAs(298), default_GaAs(298)])
    # print(my_cell)

    from solcore.poisson_drift_diffusion.DriftDiffusionUtilities import solve_pdd, default_photon_flux, \
        default_wavelengths
    import matplotlib.pyplot as plt

    solve_pdd(my_cell, 'QE', vfin=1.2, vstep=0.05, light=True)
    QE = my_cell(0).qe

    # Finally, we plot the internal and external quantum efficiencies using the information stored in the output dictionaries
    plt.plot(QE['QE']['wavelengths'] / 1e-9, QE['QE']['IQE'] * 100, label='IQE')
    plt.plot(QE['QE']['wavelengths'] / 1e-9, QE['QE']['EQE'] * 100, label='EQE')
    plt.plot(QE['QE']['wavelengths'] / 1e-9, my_cell.T * 100, label='T')
    plt.ylim(0, 100)
    plt.legend(loc='lower left')
    plt.ylabel('QE (%)')
    plt.xlabel('Wavelength (nm)')

    plt.show()



import tempfile
from copy import copy

from matplotlib import font_manager
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.backends.backend_pdf import FigureCanvasPdf
from matplotlib.colors import Normalize  # ColorConverter
import matplotlib.cm as cmx
from typing import List, Tuple, Dict, Union, Any
from .graph_support import flatten, open_with_os

graph_defaults: Dict[str, Any] = {
    "format": "pdf",
    "linetype": "-",
    "dpi": 100,
    "legendframe": True,
    "font_family": "serif",
    "palette": "bright hue wheel",
    "square": True,
    "colormap": "nipy_spectral",
    "grid": {"dash_capstyle": "round", "zorder": 1, "alpha": 0.5}
}

mpl_simple_properties: Tuple[str, ...] = (
"title", "xlabel", "ylabel", "zlabel", "xlim", "ylim", "zlim3d", "xticks", "xticklabels", "yticks", "yticklabels",
"yscale", "xscale",)
directions: Tuple[str, ...] = ("left", "right", "top", "bottom")


def make_square(fig: Figure, ax: Axes) -> None:
    bb: Any = ax.get_position()
    fwidth: float = fig.get_figwidth()
    fheight: float = fig.get_figheight()
    axwidth: float = fwidth * (bb.x1 - bb.x0)
    axheight: float = fheight * (bb.y1 - bb.y0)

    # square_edge = min((axwidth, axheight))
    if axwidth > axheight:
        narrow_by: float = (axwidth - axheight) / fwidth
        bb.x0 += narrow_by / 2
        bb.x1 -= narrow_by / 2
    elif axheight > axwidth:
        shrink_by: float = (axheight - axwidth) / fheight
        bb.y0 += shrink_by / 2
        bb.y1 -= shrink_by / 2
    ax.set_position(bb)


class Graph:
    suppress_show: bool = False
    plotted = 0

    def __init__(self, *data: Any, **options: Any) -> None:
        self.axis: Any = None
        self.options = copy(graph_defaults)
        self.options.update(options)
        self.data = list(flatten(data))
        self.extra_artists: List = []
        self.subplots: List = []
        self.create_figure()

    def main_drawing_flow(self, figure: Figure) -> None:
        # print ("mdf")
        for subplot in self.subplots:
            subplot.main_drawing_flow(figure)
        self.create_axis(figure)
        self.setup_matplotlib()
        self.render_data()
        # if "yscale" in self.options and "xscale" in self.options and self.options["xscale"] == self.options["yscale"]:
        #     self.axis.set_aspect(1./self.axis.get_data_ratio())
        # else:
        #     print ("cannot guarantee square log-lin plots yet")

        if "legend" in self.options:
            self.draw_legend()
            self.extra_artists.append(self.legend)

        if "text" in self.options:
            self.extra_artists.append(self.text)

        if "square" in self.options and self.options["square"]:
            make_square(self.figure, self.axis)

    def add_subplot(self, *data: Any, **options: Any) -> None:
        self.subplots.append(Graph(*data, figure=self.figure, **options))

    def create_figure(self):
        if "figure" in self.options:
            self.figure = self.options["figure"]
        elif "figsize" in self.options:
            self.figure = Figure(figsize=self.options["figsize"])
        else:
            self.figure = Figure()
        self.figure.subplots_adjust(**{k: self.options[k] for k in self.options.keys() if k in directions})

    def create_axis(self, figure: Figure) -> None:
        if "axis" in self.options:
            self.axis = self.options["axis"]
        elif self.axis is not None:
            # print ("already has axis")
            return
        else:
            subplot_args = self.options["subplot_args"] if "subplot_args" in self.options else {}
            self.axis = figure.add_subplot(
                self.options["subplot"] if "subplot" in self.options else 111,
                **subplot_args
            )

    def setup_matplotlib(self) -> None:
        simple_setters = {k: self.options[k] for k in mpl_simple_properties if k in self.options}
        for key, value in simple_setters.items():
            getattr(self.axis, "set_{}".format(key))(value)

        if "grid" in self.options and self.options["grid"] is not None:
            self.axis.grid(**self.options["grid"])

        if "xhide" in self.options: self.axis.get_xaxis().set_visible(False)
        if "yhide" in self.options: self.axis.get_yaxis().set_visible(False)
        if "xlabelcolor" in self.options: self.axis.get_xaxis().label.set_color(self.options["xlabelcolor"])
        if "ylabelcolor" in self.options: self.axis.get_yaxis().label.set_color(self.options["ylabelcolor"])
        if "x_major_formatter" in self.options: self.axis.get_xaxis().set_major_formatter(
            self.options["x_major_formatter"])
        if "y_major_formatter" in self.options: self.axis.get_yaxis().set_major_formatter(
            self.options["y_major_formatter"])

        # if "xticks" in self.options:
        #     x_coordinate, x_label= list(zip(*self.options["xticks"]))
        #     self.axis.set_xticks(x_coordinate)
        #     self.axis.set_xticklabels(x_label)

        if "yticklabelcolor" in self.options and self.options["yticklabelcolor"] is not None:
            for tl in self.axis.get_yticklabels():
                tl.set_color(self.options["yticklabelcolor"])
        if "xticklabelcolor" in self.options and self.options["xticklabelcolor"] is not None:
            for tl in self.axis.get_xticklabels():
                tl.set_color(self.options["xticklabelcolor"])

    def render_data(self) -> None:

        all_color_indexes = [d.color_index if (
            hasattr(d, "color_index")
            and d.color_index is not None
            and not ("color" in d.line_format)
        ) else -i - 1 for i, d in enumerate(self.data) if not (
            "color" in d.line_format or
            "facecolor" in d.line_format or
            "edgecolor" in d.line_format)]

        unique_color_indexes = len(set(all_color_indexes))

        cNorm = Normalize(vmin=0, vmax=unique_color_indexes)
        scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=self.options["colormap"])

        self.legend_additions: List = []

        for i, datum in enumerate(self.data):
            if "color" not in datum.line_format and "edgecolor" not in datum.line_format and "facecolor" not in datum.line_format:
                if hasattr(datum, "color_index") and datum.color_index is not None:
                    color = scalarMap.to_rgba(datum.color_index)
                else:
                    color = scalarMap.to_rgba(i)

                datum.line_format["color"] = color

            if "edit" in self.options:
                datum.edit = self.options["edit"]
            legend_entry = datum.draw(self.axis)
            if legend_entry is not None:
                self.legend_additions.append((i, legend_entry[0], legend_entry[1])
                                             )

    def draw_legend(self) -> None:
        font_args = {}

        if "legend_font_size" in self.options:
            font_args["size"] = self.options["legend_font_size"]
        if "legend_font_name" in self.options:
            font_args["fontname"] = self.options["legend_font_name"]
        if "legend_font_weight" in self.options:
            font_args["weight"] = self.options["legend_font_weight"]
        if "legend_font_color" in self.options:
            font_args["color"] = self.options["legend_font_color"]

        self.legend_font_properties = font_manager.FontProperties(**font_args)

        legend_location = self.options["legend"]
        handles, labels = self.axis.get_legend_handles_labels()
        self.legend_additions.sort()

        for i, (position, handle, label) in enumerate(self.legend_additions):
            handles.insert(position, handle)
            labels.insert(position, label)

        final_handles, final_labels = [], []
        for handle, label in zip(handles, labels):
            if label in ("None", None):
                continue
            final_labels.append(label)
            final_handles.append(handle)

        legend_args = (final_handles, final_labels)

        legend_kwargs = {
            "frameon": self.options["legendframe"] if "legendframe" in self.options else False,
            "title": self.options["legendtitle"] if "legendtitle" in self.options else False,
            "ncol": self.options["legendcolumns"] if "legendcolumns" in self.options else 1,
            "numpoints": 4,
            "prop": self.legend_font_properties,
            "handlelength": self.options["handlelength"] if "handlelength" in self.options else 2
        }
        if legend_location == "outside":
            legend_kwargs['bbox_to_anchor'] = self.options["legendbox"] if "legendbox" in self.options else (1.6, 1)
        else:
            legend_kwargs["loc"] = legend_location

        self.legend = self.axis.legend(*legend_args, **legend_kwargs)

    def draw(self, path: str = None, show: bool = True) -> None:
        self.main_drawing_flow(self.figure)

        if path is None:
            handle, path = tempfile.mkstemp(prefix="tmp_solcore_", suffix=".%s" % self.options["format"])

        if self.options["format"].upper() == "PDF":
            self.canvas = FigureCanvasPdf(self.figure)
        else:
            self.canvas = FigureCanvasAgg(self.figure)

        self.canvas.print_figure(path, dpi=self.options["dpi"], bbox_extra_artists=self.extra_artists,
                                 bbox_inches='tight')

        if show and not self.suppress_show:
            open_with_os(path)

    def show(self) -> None:
        import pylab
        self.main_drawing_flow(pylab.gcf())
        pylab.show()


class TwinnedAxisGraph(Graph):
    def __init__(self, *args: Any, twin:str = "x", **kwargs: Any) -> None:
        self.twin = twin
        self.line_format: Any = {}
        Graph.__init__(self, *args, **kwargs)

    def draw(self, axis: Axes) -> None:
        # print ("Td")
        if self.twin == "x":
            self.axis = axis.twinx()
        elif self.twin == "y":
            self.axis = axis.twiny()
        else:
            raise ValueError("Incorrect value for twinning_axis, must be x or y")
        self.main_drawing_flow(self.figure)



from pytest import fixture, raises
from pathlib import Path


@fixture
def user_config(tmpdir):
    return tmpdir / "solcore_config.txt"


@fixture
def default_config():
    return Path(__file__).parent.parent / "solcore" / "solcore_config.txt"


@fixture
def config(default_config, user_config):
    from solcore.config_tools import SolcoreConfig
    return SolcoreConfig(default_config, user_config)


def test_solcore_config_class(config):
    from solcore import SOLCORE_ROOT
    from random import choice
    import os

    assert os.path.isfile(config.user_config)
    assert config.sections == config._default_data.sections()

    s = choice(config.sections)
    o = choice(config.sources(s))
    assert config[s, o] == config._user_data[s][o].replace("SOLCORE_ROOT", SOLCORE_ROOT)
    assert config[s] == config._user_data[s]
    with raises(KeyError):
        assert config["cat"] == config._user_data[s]


def test_add_remove_source(config):
    from random import choice
    from configparser import ConfigParser

    # Added correctly
    s = choice(config.sections)
    config[s, "cat"] = "/kittens/pictures"
    assert "cat" in config.sources(s)
    uconf = ConfigParser()
    uconf.read(config.user_config)
    assert "cat" in uconf.options(s)

    # Removed correctly
    config.remove_source(s, "cat")
    assert "cat" not in config.sources(s)
    uconf = ConfigParser()
    uconf.read(config.user_config)
    assert "cat" not in uconf.options(s)

    # Removing a default Solcore source: it is disabled, instead
    o = choice(config.sources(s))
    config.remove_source(s, o)
    assert config[s, o] == ""
    uconf = ConfigParser()
    uconf.read(config.user_config)
    assert uconf[s][o] == ""


def test_restore_default_source(config):
    from random import choice

    s = choice(config.sections)
    o = choice(config.sources(s))
    value = config[s, o]

    # Wrong source: value unchanged
    config[s, "night"] = "owl"
    assert value == config[s, o]

    # Correct source: value changed
    config[s, o] = "owl"
    assert value != config[s, o]

    # Restored
    config.restore_default_source(s, o)
    assert value == config[s, o]


def test_restore_defaults(config):
    from random import choice

    s = choice(config.sections)
    o = choice(config.sources(s))
    value = config[s, o]

    config[s, "night"] = "owl"
    config[s, o] = "owl"
    assert config[s, "night"] == "owl"
    assert value != config[s, o]

    config.restore_defaults()
    assert config[s, "night"] == "owl"
    assert value == config[s, o]


def test_register_observer(config):
    from random import choice

    def dummy():
        pass

    s = choice(config.sections)
    config.register_observer(s, dummy)

    assert s in config._observers
    assert len(config._observers[s]) == 1
    assert config._observers[s][0] == dummy


def test_notify_observer(config):
    from random import choice

    sources = {}

    def dummy(source, value):
        sources[source] = value

    s = choice(config.sections)
    o = choice(config.sources(s))
    config.register_observer(s, dummy)

    config[s, o] = "spider"
    assert sources[o] == config[s, o]


def test_welcome_message(config):
    config.welcome_message(False)
    assert config.welcome_message() is False


def test_verbose_loading(config):
    config.verbose_loading(False)
    assert config.verbose_loading() is False


def test_spice(config):
    config.spice("neither/here/nor/there")
    assert config.spice() == "neither/here/nor/there"


def test_smarts(config):
    config.smarts("neither/here/nor/there")
    assert config.smarts() == "neither/here/nor/there"


def test_units(config):
    from random import choice

    assert config.units() == config.sources("Units")
    o = choice(config.sources("Units"))
    config.units(o, "neither/here/nor/there")
    assert config.units(o) == "neither/here/nor/there"


def test_parameters(config):
    from random import choice

    assert config.parameters() == config.sources("Parameters")
    o = choice(config.sources("Parameters"))
    config.parameters(o, "neither/here/nor/there")
    assert config.parameters(o) == "neither/here/nor/there"


def test_materials(config):
    from random import choice

    assert config.materials() == config.sources("Materials")
    o = choice(config.sources("Materials"))
    config.materials(o, "neither/here/nor/there")
    assert config.materials(o) == "neither/here/nor/there"



import sys
from random import random

import numpy as np

from solcore.constants import *
from solcore.smooth import smooth
from solcore.structure import *
from solcore.quantum_mechanics.heterostructure_alignment import VBO_align
from solcore.quantum_mechanics.strain import strain_calculation_parameters

from solcore.quantum_mechanics.kp_QW import kp6x6, kp4x4
from solcore.quantum_mechanics.kp_bulk import kp8x8_bulk

m0 = electron_mass


def assemble_qw_structure(repeats, well, bulk_l_top, bulk_l_bottom, barrier, well_interlayer=None,
                          structure_label="QW Structure", shift_wells=0):
    half_barrier = Layer(barrier.width / 2, barrier.material)
    qw_structure = Structure()
    # qw_structure.append(bulk, layer_label="bulk")
    qw_structure.append_multiple([bulk_l_top], layer_labels=["bulk"])
    qw_structure.append_multiple([half_barrier], layer_labels=["half barrier"])
    if well_interlayer:
        qw_structure.append_multiple([well_interlayer, well, well_interlayer, barrier], layer_labels=["interlayer",
                                                                                                      "well",
                                                                                                      "interlayer",
                                                                                                      "barrier"],
                                     repeats=repeats - 1)
        qw_structure.append_multiple([well_interlayer, well, well_interlayer, half_barrier, bulk_l_bottom],
                                     layer_labels=[
                                         "interlayer", "well", "interlayer", "half barrier", "bulk"])
    else:
        qw_structure.append_multiple([well, barrier], layer_labels=["well", "barrier"], repeats=repeats - 1)
        qw_structure.append_multiple([well, half_barrier, bulk_l_bottom], layer_labels=["well", "half barrier", "bulk"])

    return qw_structure


def vary_well_widths(structure, fraction=0.5, region_sought="well"):
    """Returns an array of indices which correspond to the region that are quantum wells in the structure."""
    depth = 0

    for i, (layer, label) in enumerate(zip(structure, structure.labels)):
        if label != region_sought:
            continue

        layer.width *= (1 + (2 * random() - 1) * fraction)
        print(layer.width)
    return structure


def locate_regions(x, structure, region_sought="well"):
    """Returns an array of indices which correspond to the region that are quantum wells in the structure."""
    depth = 0

    well_structure_indices = set([item for item in range(
        len(structure.labels)) if structure.labels[
                                      item] == region_sought])  # <---- Uses new labels in the structure object
    well_z_indices = np.array([], dtype=np.dtype(int))
    for i, layer in enumerate(structure):
        if well_structure_indices.issuperset([i]):
            well_z_indices = np.hstack((well_z_indices, np.where((depth <= x) * (x <= depth + layer.width) == True)[0]))
        depth += layer.width
    return well_z_indices


def well_regions(x, structure):
    return locate_regions(x, structure, "well")


def text_render(structure, resolution=100):
    x = np.linspace(0, structure.width(), resolution)
    bulk = locate_regions(x, structure, "bulk")
    barrier = set(locate_regions(x, structure, "barrier")) | set(locate_regions(x, structure, "half barrier"))
    interlayer = locate_regions(x, structure, "interlayer")
    well = locate_regions(x, structure, "well")

    chars = ["―" if i in bulk else "_" if i in well else "-" if i in interlayer else "‾" if i in barrier else "?" for i
             in range(resolution)]
    return ("".join(chars))


def structure_to_potentials(structure, step_size=None, minimum_step_size=0, smallest_feature_steps=20,
                            blur=False, blurmode="even", return_qw_boolean_for_layer=False, Efield=0,
                            mode='kp4x4'):
    """ Discretizes the structure as a function of the position, providing the potential for the electrons, HH, LH and
    SO bands as well as the Luttinger parameters and the effective masses.

    The result depend on the chosen mode:
        - kp4x4, calculates the bands and effective masses at k=0 coupling the HH and LH bands
        - kp6x6, calculates the bands and effective masses at k=0 coupling the HH, LH and SO bands
        - kp8x8_bulk, calculates the four bands and fit the effective masses with a parabola around k=0
        - strain, just shifts the bands according to the strain
        - relaxed, do nothing and things are calculated as if we had the bulk, unstrained materials

    Notice that although the kp8x8_bulk couples the 4 (8) bands, the effective mass is the result of a fitting around
    k=0. Therefore, it is not valid to solve 1D problems, but just a bulk-like problem under a parabolic aproximation

    Additionlly, the band structure can be blured - to simulate the intermixing of materials - as well as being under
    an electric field.
    """

    # We calculate the mesh based on the minimum feature of the structure and the number of points desired for that
    # feature

    available_modes = ['kp6x6', 'kp4x4', 'kp8x8_bulk', 'strain', 'relaxed']
    assert mode in available_modes, 'ERROR: Calculation mode must be {}'.format(available_modes)


    widths = [layer.width for layer in structure]
    total_width = sum(widths)
    if step_size is None:
        smallest_feature_width = min(widths)
        step_size = int(smallest_feature_steps / smallest_feature_width * total_width)
        step_size = max(step_size, minimum_step_size)

    x = np.linspace(0, total_width, step_size)
    Ve = np.zeros(len(x))  # Electron potential
    Vhh = np.zeros(len(x))  # Heavy hole    "
    Vlh = np.zeros(len(x))  # Light hole    "
    Vso = np.zeros(len(x))  # SO band    "
    me = np.zeros(len(x))  # Electron effective mass
    mhh_p = np.zeros(len(x))  # HH paralell effective mass
    mhh_t = np.zeros(len(x))  # HH transverse effective mass
    mlh_p = np.zeros(len(x))  # LH paralell effective mass
    mlh_t = np.zeros(len(x))  # LH transverse effective mass
    g1 = np.zeros(len(x))  # Gamma 1
    g2 = np.zeros(len(x))  # Gamma 2
    g3 = np.zeros(len(x))  # Gamma 3
    isWell = np.zeros(len(x), dtype=bool)  # indicate where we have the QW    "

    # Spatially varying effective mass
    depth = 0

    # At each position of the structure, we have to calculate the effective mass of electrons and holes, and the band
    # edges.
    for i, layer in enumerate(structure):

        positions = (depth <= x) * (x <= depth + layer.width)

        me[positions] = layer.material.get("eff_mass_electron") * electron_mass
        g1[positions] = layer.material.get("gamma1")
        g2[positions] = layer.material.get("gamma2")
        g3[positions] = layer.material.get("gamma3")
        mhh_p[positions] = 1. / (g1[positions] - 2 * g2[positions]) * m0
        mhh_t[positions] = 1. / (g1[positions] + 2 * g2[positions]) * m0
        mlh_p[positions] = 1. / (g1[positions] + g2[positions]) * m0
        mlh_t[positions] = 1. / (g1[positions] - g2[positions]) * m0

        if return_qw_boolean_for_layer:
            isWell[positions] = (layer == return_qw_boolean_for_layer)

        try:
            Ve[positions] = layer.Ec
            Vhh[positions] = layer.Ev
            Vlh[positions] = layer.Ev
            Vso[positions] = layer.Ev - layer.material.spin_orbit_splitting

        except KeyError:
            print("There was a problem converting the structure to a potential because some information was missing "
                  "from the structure state.")
            print("Try calling the function 'align_heterostructure_using_Vurgaftman' before calling "
                  "'structure_to_potentials'.")
            sys.exit()

        except Exception as inst:
            print(inst)
            raise

        substrate = structure.substrate

        # Now we modify the band profile and effective masses according to the chosen mode:
        # - kp4x4, calculates the bands and effective masses at k=0 coupling the HH and LH bands
        # - kp6x6, calculates the bands and effective masses at k=0 coupling the HH, LH and SO bands
        # - kp8x8_bulk, calculates the four bands and fit the effective masses with a parabola around k=0
        # - strain, just shifts the bands according to the strain
        # - relaxed, do nothing and things are calculated as if we had the bulk, unstrained materials
        if mode == 'kp6x6':
            # The band edges are shifted and the effective mass of the LHs also changes.
            c, hh, lh, so, mc_kp, mhh_p_kp, mlh_p_kp, mso_p_kp, mhh_t_kp, mlh_t_kp, mso_t_kp = kp6x6(layer.material,
                                                                                                         substrate)

            mlh_p[positions] = mlh_p_kp
            mlh_t[positions] = mlh_t_kp

            Ve[positions] = c
            Vhh[positions] = hh
            Vlh[positions] = lh
            Vso[positions] = so

        elif mode == 'kp4x4':
            # The band edges are shifted but the effective masses around k=0 are not affected.
            c, hh, lh, mc_kp, mhh_p_kp, mlh_p_kp, mhh_t_kp, mlh_t_kp = kp4x4(layer.material, substrate)

            Ve[positions] = c
            Vhh[positions] = hh
            Vlh[positions] = lh

        elif mode == 'kp8x8_bulk':
            # The band edges and the effective mass of Electrons, HH, LH and the SO holes changes.
            c, hh, lh, so, mc, mhh, mlh, mso = kp8x8_bulk(layer.material, substrate)

            me[positions] = mc
            mhh_p[positions] = mhh
            mlh_p[positions] = mlh
            mhh_t[positions] = mhh
            mlh_t[positions] = mlh

            Ve[positions] = c
            Vhh[positions] = hh
            Vlh[positions] = lh
            Vso[positions] = so

        elif mode == 'strain':
            # The bands are just shifted according to strain
            strain_parameters = strain_calculation_parameters(substrate, layer.material, SO=True)
            Ve[positions] = Ve[positions] + strain_parameters.delta_Ec
            Vhh[positions] = Vhh[positions] + strain_parameters.delta_Ehh
            Vlh[positions] = Vlh[positions] + strain_parameters.delta_Elh

        else:
            pass

        depth += layer.width

    # We blur the structure if required
    if blur:
        Ve = smooth(Ve, blur / total_width * step_size, blurmode=blurmode)
        Vhh = smooth(Vhh, blur / total_width * step_size, blurmode=blurmode)
        Vlh = smooth(Vlh, blur / total_width * step_size, blurmode=blurmode)
        Vso = smooth(Vso, blur / total_width * step_size, blurmode=blurmode)

        me = smooth(me, blur / total_width * step_size, blurmode=blurmode)
        g1 = smooth(g1, blur / total_width * step_size, blurmode=blurmode)
        g2 = smooth(g2, blur / total_width * step_size, blurmode=blurmode)
        g3 = smooth(g3, blur / total_width * step_size, blurmode=blurmode)

        mhh_p = smooth(mhh_p, blur / total_width * step_size, blurmode=blurmode)
        mhh_t = smooth(mhh_t, blur / total_width * step_size, blurmode=blurmode)
        mlh_p = smooth(mlh_p, blur / total_width * step_size, blurmode=blurmode)
        mlh_t = smooth(mlh_t, blur / total_width * step_size, blurmode=blurmode)

    # We tilt the structure if there is an electric field
    if Efield != 0:
        Ve = Ve + q * x * Efield
        Vhh = Vhh + q * x * Efield
        Vlh = Vlh + q * x * Efield
        Vso = Vso + q * x * Efield

    if mode in ['kp4x4', 'strain', 'relaxed']:
        Vso = Vso * 0


    if mode in ['kp4x4', 'kp6x6']:
        return {"x": x,
                "Ve": Ve,
                "me": me,
                "Vhh": Vhh,
                "g1": g1,
                "Vlh": Vlh,
                "g2": g2,
                "Vso": Vso,
                "g3": g3,
                "isWell": isWell,
                "mhh_p": mhh_p,
                "mhh_t": mhh_t,
                "mlh_p": mlh_p,
                "mlh_t": mlh_t}
    else:
        return {"x": x,
                "Ve": Ve,
                "me": me,
                "Vhh": Vhh,
                "Vlh": Vlh,
                "Vso": Vso,
                "isWell": isWell,
                "mhh": mhh_p,
                "mlh": mlh_p}


if __name__ == "__main__":
    from solcore import si, material
    from solcore.structure import Layer
    import matplotlib.pyplot as plt

    bulk = material("GaAs")(T=293)
    QW = material("InGaAs")(T=293, In=0.147)
    barrier = material("GaAsP")(T=293, P=0.1)

    bulk.strained = False
    QW.strained = True
    barrier.strained = True

    top_layer = Layer(width=si("50nm"), material=bulk)
    well_layer = Layer(width=si("7.2nm"), material=QW)
    barrier_layer = Layer(width=si("29nm"), material=barrier)
    bottom_layer = top_layer

    print("Here is a ASC_examples structure with no materials:")

    test_structure = assemble_qw_structure(
        repeats=3,
        well=well_layer,
        bulk_l_top=top_layer,
        bulk_l_bottom=bottom_layer,
        barrier=barrier_layer,
    )

    test_structure.substrate = bulk

    VBO_align(test_structure)
    result = structure_to_potentials(test_structure, mode='kp8x8_bulk')
    print(test_structure)

    # plt.plot(result['x']*1e9, result['Ve']/q, result['x']*1e9, result['Vhh']/q, result['x']*1e9, result['Vlh']/q, result['x']*1e9, result['Vso']/q)
    #    plt.ylim(-1.5, 1.0)
    # plt.plot(result['x']*1e9, result['me']/m0, result['x']*1e9, result['mhh_p']/m0, result['x']*1e9, result['mlh_p']/m0)
    # plt.plot(result['x'] * 1e9, result['me'] / m0, result['x'] * 1e9, result['mhh_t'] / m0, result['x'] * 1e9,
    #          result['mlh_t'] / m0)

    plt.show()



""" This module contains a collection of mathematical models to calculate the dielectric constant of a material. The
modelling is largely based in the equations and naming used by:

J. A. Woollam, Guide to using WVASE: Spectroscopic Ellipsometry Data Acquisition and Analysis Software. 2012, pp. 1–696.

"""
import numpy as np
from solcore.absorption_calculator.kramers_kronig import Epsi1
from solcore.interpolate import interp1d


class Poles:
    """ Basic oscillator model for fitting ellipsometry and RAT data. It a pole without broadening and therefore can
    only represent non-absorbing materials. Usually, the central energy must lie outside the spectral range of interest.

    .. math:: \\epsilon = \\frac {A_n E_n^2} {E_n^2 - E^2}

    with:

    - An = the amplitud of the oscillator (dimensionless)
    - En = the central energy of the pole (in eV)

    It is equivalent to Pol.2 in WVASE

    """
    name = 'poles'
    var = 2

    def __init__(self, A=0., Ec=5.):
        """ Constructor of the pole oscillator. By default, the amplitudes are set to zero meaning that there are
        no poles.

        :param An: Amplitud of the pole (dimensionless)
        :param En: Center energy of the pole (must lay outside the spectral range of interest (eV)).
        """
        self.A = A
        self.Ec = Ec

    def dielectric(self, x):
        """ Returns the dielectric function as modelled by this class at the given x input values (a single value or
        an array of values.

        :param x: Spectral position in nm
        :return: The dielectric constant at this spectral position
        """
        x_eV = 1240 / x  # We transform it into eV

        return self.A * self.Ec ** 2 / (self.Ec ** 2 - x_eV ** 2)

    def __repr__(self):
        out = 'Poles oscillator with:\n\t- An = {}\n\t- En = {} eV\n'.format(self.A, self.Ec)

        return out


class Lorentz:
    """ Classic Lorentz oscillator involving a central energy and a broadening responsible for the absorption.

    .. math:: \\epsilon = \\frac {A_n E_n^2} {E_n^2 - E^2 - i Br_n E}

    with:

    - An = the amplitud of the oscillator (dimensionless)
    - Brn = the broadening (eV)
    - En = the central energy (eV)

    It is equivalent to Lor.2 in WVASE.

    """
    name = 'lorentz'
    var = 3

    def __init__(self, An=0., En=2., Brn=0.1):
        """ Constructor of the Lorentz oscillator. By default, the amplitude and broadening are set to zero meaning
        that there are no oscillator.

        :param An: Amplitud of the oscillator (dimensionless)
        :param En: Center energy of the oscillator (in eV).
        :param Brn: Broadening (in eV)
        """

        self.An = An
        self.En = En
        self.Brn = Brn

    def dielectric(self, x):
        """ Returns the dielectric function as modelled by this class at the given x input values (a single value or
        an array of values.

        :param x: Spectral position in nm
        :return: The dielectric constant at this spectral position
        """
        x_eV = 1240 / x
        return self.An * self.En ** 2 / (self.En ** 2 - x_eV ** 2 - 1.j * self.Brn * x_eV)

    def __repr__(self):
        out = 'Lorentz oscillator with:\n\t- An = {}\n\t- En = {} eV\n\t- Brn = {} eV\n'.format(self.An, self.En,
                                                                                                self.Brn)

        return out


class Gauss:
    """ Gauss oscillator involving a central energy and a broadening responsible for the absorption. It is defined as:

    .. math:: \\epsilon = \\epsilon_1 + i \\epsilon_2
    .. math:: \\epsilon_2 = A \\exp^{-{\\left( \\frac{E - E_c}{\\sigma} \\right)}^2} - A \\exp^{-{\\left( \\frac{E + E_c}{\\sigma} \\right)}^2}
    .. math:: \\sigma = \\frac{Br}{2 \\sqrt{\\ln(2)}}

    Epsilon_1 is calculated using the Kramers-Kronig relationships

    with:

    - A = the amplitud of the oscillator (dimensionless)
    - Br = the broadening (eV)
    - Ec = the central energy (eV)

    It is equivalent to Gau.0 in WVASE.

    """
    name = 'gauss'
    var = 3

    def __init__(self, A=0., Ec=2., Br=0.1):
        """ Constructor of the Gauss oscillator. By default, the amplitude and broadening are set to zero meaning
        that there are no oscillator.

        :param A: Amplitud of the oscillator (dimensionless)
        :param Ec: Center energy of the oscillator (in eV).
        :param Br: Broadening (in eV)
        """

        self.A = A
        self.Ec = Ec
        self.Br = Br

    def dielectric(self, x):
        """ Returns the dielectric function as modelled by this class at the given x input values (a single value or
        an array of values.

        :param x: Spectral position in nm
        :return: The dielectric constant at this spectral position
        """
        x_eV = 1240 / x
        s = 0.5 * self.Br / np.sqrt(np.log(2))

        epsi2 = lambda xx: self.A * (np.exp(-((xx - self.Ec) / s) ** 2) - np.exp(-((xx + self.Ec) / s) ** 2))

        epsi1 = Epsi1(epsi2)

        e1 = epsi1(x_eV)
        e2 = epsi2(x_eV)

        return e1 + 1.j * e2

    def __repr__(self):
        out = 'Gauss oscillator with:\n\t- A = {}\n\t- Ec = {} eV\n\t- Br = {} eV\n'.format(self.A, self.Ec, self.Br)

        return out


class Drude:
    """ The classic Drude oscillator describes free carrier effects on the dielectric response. Its form is a Lorentz
    oscillator with zero center energy.

    .. math:: \\epsilon = - \\frac {A_n Br_n} {E^2 + i Br_n E}

    with:

    - An = the amplitud of the oscillator (eV)
    - Brn = the broadening (eV)

    It is equivalent to Drd.0 in WVASE.

    """
    name = 'drude'
    var = 2

    def __init__(self, An=0., Brn=0.1):
        """ Constructor of the Drude oscillator. By default, the amplitude is set to zero meaning that there are no
        oscillator.

        :param An: Amplitud of the oscillator (in eV)
        :param Brn: Broadening (in eV)
        """

        self.An = An
        self.Brn = Brn

    def dielectric(self, x):
        """ Returns the dielectric function as modelled by this class at the given x input values (a single value or
        an array of values.

        :param x: Spectral position in nm
        :return: The dielectric constant at this spectral position
        """
        x_eV = 1240 / x
        epsi = - self.An * self.Brn / (x_eV ** 2 + 1.j * self.Brn * x_eV)

        return epsi

    def __repr__(self):
        out = 'Drude oscillator with:\n\t- An = {} eV\n\t- Brn = {} eV\n'.format(self.An, self.Brn)

        return out


class Cauchy:
    """ Cauchy oscillator model for fitting ellipsometry data. It is generally used to fit non absorbing materials,
    although an Urbach tail is included to indicate the onset of absorption.

    .. math:: N = An + \\frac {Bn} {E^2} + \\frac {Cn} {E^4}
    .. math:: K = A_k \\exp(B_k (E-C_k))
    .. math:: \\epsilon = (N + iK)^2

    with:

    - An, Bn and Cn = the Cauchy coeficients (dimensionless, µm-2 and µm-4, respectively)
    - Ak = the amplitud of the Urbach tail (dimensionless)
    - Bk = the multiplicative factor of the Urbach tail (eV-1)
    - Ck = the energy offset of the Urbach tail (eV) - it is user selected, but redundant with Ak.
    with Ak.

    It is equivalent to Chy.0 in WVASE.

    """
    name = 'cauchy'
    var = 5

    def __init__(self, An=0., Bn=0., Cn=0., Ak=0., Bk=0., Ck=0):
        """ Constructor of the Cauchy oscillator.

        :param An: 1st Cauchy coefficient (dimensionless)
        :param Bn: 2nd Cauchy coefficient (µm-2)
        :param Cn: 3rd Cauchy coefficient (µm-4)
        :param Ak: Amplitude of the Urback tail (dimensionless)
        :param Bk: multiplicative factor of the Urbach tail (eV-1)
        :param Ck: energy offset (eV)
        """

        self.An = An
        self.Bn = Bn
        self.Cn = Cn
        self.Ak = Ak
        self.Bk = Bk
        self.Ck = Ck

    def dielectric(self, x):
        """ Returns the dielectric function as modelled by this class at the given x input values (a single value or
        an array of values.

        :param x: Spectral position in nm
        :return: The dielectric constant at this spectral position
        """
        x_eV = 1240 / x

        N = self.An + self.Bn / (x / 1000.) ** 2 + self.Cn * (x / 1000.) ** 4
        K = self.Ak * np.exp(self.Bk * (x_eV - self.Ck))

        return (N + 1.j * K) ** 2

    def __repr__(self):
        out = 'Cauchy oscillator with:\n\t- An = {}\n\t- Bn = {} µm-2\n\t- Cn = {} µm-4\n\t- Ak = {}\n\t' \
              '- Bk = {} eV-1\n\t- Ck = {} eV\n'.format(self.An, self.Bn, self.Cn, self.Ak, self.Bk, self.Ck)

        return out


class PolySegment:
    """ The PolySegment model aims to approximate the dielectric function with a finite number of linear segments. It is
    similar to a point-by-point fitting but likely using much less free variables. The segments are created for
    Epsilon_2 and Epsilon_1 is calculated using the Kramers-Kronig relationships.

    It allows for a maximum freedom for fitting data while reducing noise, but does not provide any physically
    relevant information.

    The input parameters are the energies of the endpoints of the segments and the corresponding Epsilon_2 values.

    """
    name = 'polysegment'

    def __init__(self, energy, e2):
        """ Constructor of the PolySegment oscillator. By default, the amplitude and broadening are set to zero meaning
        that there are no oscillator.

        :param energy: Energies (eV)
        :param e2: Epsilon_2 values at the corresponding energies (dimensionless)
        """

        self.energy = energy
        self.e2 = e2
        self.epsi2 = interp1d(x=energy, y=e2)
        self.var = len(e2)

    def dielectric(self, x):
        """ Returns the dielectric function as modelled by this class at the given x input values (a single value or
        an array of values.

        :param x: Spectral position in nm
        :return: The dielectric constant at this spectral position
        """
        x_eV = 1240 / x

        epsi2 = lambda xx: self.epsi2(x_eV)

        epsi1 = Epsi1(epsi2)

        e1 = epsi1(x_eV)
        e2 = epsi2(x_eV)

        return e1 + 1.j * e2

    def __repr__(self):
        out = 'PolySegment oscillator with endpoints:\n' \
              '\t- Emin = {}\n' \
              '\t- Emax = {} eV\n' \
              '\t- Epsilon2(Emin) = {}\n' \
              '\t- Epsilon2(Emax) = {}\n'.format(self.energy[0], self.energy[-1], self.e2[0], self.e2[-1])

        return out


# Oscillator class designed to build an optical constrant model structure for customised modelling...
class Oscillator:
    def __init__(self, oscillator_type, material_parameters=None, **kwargs):
        self.oscillator = oscillator_type
        self.material_parameters = material_parameters
        self.arguments = kwargs

        if self.material_parameters == None:

            self.Material = "Not Specified"
        else:

            self.Material = self.material_parameters["Material"]

    def __str__(self):
        return "< Oscillator Type :: " + self.oscillator + ", Material :: " + self.Material + " >"


class DielectricConstantModel(list):
    """ This class groups together several models for the dielectric constant applicable to a given layer. Based on
    all of them, it calculates the dielectric constant and the complex refractive index at any energy. It requires as
    input an offset, offen refered as epsilon_infinity.

    """

    def __init__(self, e_inf=1., oscillators=()):
        """ Creates the dielectric constant model. The default model is air (constant dielectric constant = 1 at all
          wavelengths.

        :param e_inf: The offset dielectric constant. Default=1
        :param oscillators: A list containing the oscillators. Default=()
        """
        list.__init__(self, oscillators)

        self.e_inf = e_inf

    def add_oscillator(self, name, **kwargs):
        """ Adds an oscillator to the dielectric constant model.

        :param name: The name of the oscillator
        :param kwargs: The arguments needed to create that oscillator.
        :return:
        """
        available_models = {'poles': Poles,
                            'lorentz': Lorentz,
                            'drude': Drude,
                            'cauchy': Cauchy,
                            'gauss': Gauss,
                            'polysegment': PolySegment}

        assert name in available_models.keys(), 'Error: unknown oscillator model. ' \
                                                'Valid names are: {}'.format(available_models.keys())

        self.append(available_models[name](**kwargs))

    def remove_oscillator(self, idx):
        """ Removes an oscillator from the model.

        :param idx: The number of the oscillator (starting in 1)
        :return: None"""

        assert 0 < idx <= len(self), 'Error: The available oscillators are:\n {}'.format(self.__repr__())

        print(self[idx - 1])
        a = input('Do you want to remove this oscillator from the model (Y/n)?')

        if a in 'Yy':
            self.pop(idx - 1)

    def dielectric_constants(self, x):
        """ Returns the complex dielectric constant in the form e1 + ie2

        :param x: The spectral possition in nm (it can be an array)
        :return: The complex dielectric constant.
        """
        # If the wavelength is very small... then we are in the SI units. Lets change it to nm.
        try:
            in_si = any(x < 0.001)
        except TypeError:
            in_si = x < 0.001

        if in_si:
            x = x * 1e9

        epsi = self.e_inf
        for model in self:
            epsi = epsi + model.dielectric(x)

        return epsi

    def n_and_k(self, x):
        """ Returns the complex refractive index in the form n + ik

        :param x: The spectral possition in nm (it can be an array)
        :return: The complex refractive index.
        """

        return np.sqrt(self.dielectric_constants(x))

    def __repr__(self):

        out = 'e_inf = {}\n'.format(self.e_inf)
        for i in range(len(self)):
            out += '{} - {}'.format(i + 1, self[i].__repr__())

        return out


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from solcore.absorption_calculator.transfer_matrix import calculate_rat

    E = 2 * np.logspace(2, 4, 200)

    a = Poles(A=0, Ec=2)
    loz = Lorentz(An=1, En=2, Brn=0.5)
    drud = Drude(An=24.317000, Brn=0.125740)
    gau1 = Gauss(A=2.8468, Ec=0.1299, Br=0.0111)
    gau2 = Gauss(A=7.192, Ec=0.058331, Br=0.01682)
    gau3 = Gauss(A=1.9737, Ec=0.13991, Br=0.02144)
    c = Cauchy()

    model = DielectricConstantModel(e_inf=2.9987, oscillators=[gau1, gau2, gau3])

    # out = calculate_rat([[300, model]], x)

    # plt.semilogx(x, out['R'], 'b', label='Reflexion')
    # plt.semilogx(x, out['A'], 'r', label='Absorption')
    # plt.semilogx(x, out['T'], 'g', label='Transmission')

    n = model.n_and_k(E)
    #
    plt.semilogx(E, np.real(n), 'b', label='n')
    plt.semilogx(E, np.imag(n), 'r', label='k')
    plt.legend()
    plt.show()



"""
This module interpolates the n and k data of an alloy based on the known values at certain specific compositions.

The way it works is by interpolating in a smart way certain critical points (usually the Adachi critical points) and
then filling the gaps in between.
"""
import numpy as np
import os
from configparser import ConfigParser


def interpolate_critical_points(critical_points, target):
    fractions_available = np.array(sorted(critical_points.keys()))

    target_index = fractions_available.searchsorted(target)
    lower = fractions_available[target_index - 1]
    upper = fractions_available[target_index]

    fraction_along = (target - lower) / (upper - lower)
    lower_points = critical_points[lower]
    upper_points = critical_points[upper]

    result_critical_points = {k: lower_points[k] + fraction_along * (upper_points[k] - lower_points[k]) for k in
                              lower_points.keys()}
    return result_critical_points, (lower, upper), fraction_along


def load_data_from_directory(directory):
    result_config = ConfigParser()
    result_config.read(os.path.join(directory, "critical_points.txt"))
    sections = result_config.sections()

    fractions_config_file = [float(s) for s in sections]

    critical_points = {}
    fraction_dict = {}
    for f, s in zip(fractions_config_file, sections):
        options = sorted(result_config.options(s))

        critical_points[f] = {o: np.array([float(f) for f in result_config.get(s, o).split()]) for o in (options) if
                              o != "file"}
        datapath = os.path.join(directory, result_config.get(s, "file"))

        fraction_dict[f] = np.loadtxt(datapath, unpack=True)

    return fraction_dict, critical_points


def split_data_at_points(data, points):
    split_at = data[0].searchsorted(points[:, 0])

    split_data = (np.split(data, split_at, axis=1))
    return split_data


def transform(data, from_critical_points, to_critical_points):
    # sort the critical points according to "from" side order
    key_order, ordered_from_critical_points = zip(*sorted(from_critical_points.items(), key=lambda item: item[1][0]))
    moo, ordered_to_critical_points = zip(
        *sorted(to_critical_points.items(), key=lambda item: key_order.index(item[0])))
    ordered_to_critical_points = np.array(ordered_to_critical_points)
    ordered_from_critical_points = np.array(ordered_from_critical_points)
    split_data = split_data_at_points(data, ordered_from_critical_points)

    full_interval_bounds = np.concatenate(([0], ordered_from_critical_points[:, 0], [1e10]))
    full_interval_bounds_target = np.concatenate(([0], ordered_to_critical_points[:, 0], [1e10]))

    full_interval_bounds_y = np.concatenate(([0], ordered_from_critical_points[:, 1], [0]))
    full_interval_bounds_target_y = np.concatenate(([0], ordered_to_critical_points[:, 1], [0]))

    result_x, result_y = [], []

    for interval, data_subset in enumerate(split_data):
        interval_bounds = full_interval_bounds[interval], full_interval_bounds[interval + 1]
        interval_bounds_y = full_interval_bounds_y[interval], full_interval_bounds_y[interval + 1]

        interval_bounds_target = full_interval_bounds_target[interval], full_interval_bounds_target[interval + 1]
        interval_bounds_target_y = full_interval_bounds_target_y[interval], full_interval_bounds_target_y[interval + 1]

        if len(data_subset) == 0:
            continue
        x, y, *_ = data_subset  # heh heh heh, *_

        fraction_along = (x - interval_bounds[0]) / (interval_bounds[1] - interval_bounds[0])

        fraction_up = (y - interval_bounds_y[0]) / (interval_bounds_y[1] - interval_bounds_y[0])

        transformed_x = interval_bounds_target[0] + fraction_along * (
            interval_bounds_target[1] - interval_bounds_target[0])
        transformed_y = interval_bounds_target_y[0] + fraction_up * (
            interval_bounds_target_y[1] - interval_bounds_target_y[0])

        result_x.append(transformed_x)
        result_y.append(transformed_y)

    return np.array((np.concatenate(result_x), np.concatenate(result_y)))


def critical_point_interpolate(data, critical_points, target_fraction, grid):
    result_critical_points, (lower_fract, higher_fract), fraction_along = \
        interpolate_critical_points(critical_points, target_fraction)

    interpolate_from_lower = transform(
        data[lower_fract],
        from_critical_points=critical_points[lower_fract],
        to_critical_points=result_critical_points)
    interpolate_from_higher = transform(
        data[higher_fract],
        from_critical_points=critical_points[higher_fract],
        to_critical_points=result_critical_points)

    newgrid_lower = np.interp(grid, interpolate_from_lower[0], interpolate_from_lower[1])
    newgrid_higher = np.interp(grid, interpolate_from_higher[0], interpolate_from_higher[1])

    return grid, (newgrid_higher * fraction_along + newgrid_lower * (1 - fraction_along)), result_critical_points


if __name__ == "__main__":

    import matplotlib
    import matplotlib.pyplot as plt
    from matplotlib import animation

    # Animate only works with Tk, so we tell matpotlib to use it
    matplotlib.use("TkAgg")

    data, critical_points = load_data_from_directory(os.path.join(os.path.split(__file__)[0], 'Data', 'AlGaAs nk'))

    fig = plt.figure()

    plt.yscale("log")
    nm = np.linspace(200, 1000, 1000)
    for k in data.keys():
        # print (data[k][0]*1e9, data[k][1])
        plt.plot(data[k][0] * 1e9, data[k][1], label=str(k), color="grey")

    resultx, resulty, crit = critical_point_interpolate(data, critical_points, 0, nm * 1e-9)

    lines, = plt.plot(resultx * 1e9, resulty, label="result", color="black", linewidth=2)

    plt.xlim(0, 1000)

    crit = np.array(list(crit.values())).transpose()
    print(crit, "lr")
    markers, = plt.plot(crit[0] * 1e9, crit[1], "+", label="result", color="red", linewidth=2, markersize=10)


    def animate(i):
        f = i / 100
        resultx, resulty, crit = critical_point_interpolate(data, critical_points, f, nm * 1e-9)
        # crit = crit.transpose()
        lines.set_ydata(resulty)
        # markers.set_ydata(crit[1])
        # markers.set_xdata(crit[0]*1e9)
        fig.canvas.get_tk_widget().update()  # process events
        return lines  # , markers


    anim = animation.FuncAnimation(fig, animate, init_func=None,
                                   frames=100, interval=2, blit=False)

    # anim.save('mpp.mp4', fps=30, extra_args=['-vcodec', 'libx264'])
    w = animation.FFMpegFileWriter(fps=10)
    w.fps = 10
    # anim.save('ellipse.mp4', writer=w, fps=10, bitrate=100, extra_args=['-vcodec', 'libx264'])

    plt.show()


    #
    # print (lines)
    # for f in np.linspace(0.01,1,100):
    #     resultx, resulty, crit=critical_point_interpolate(data, critical_points, f,nm)
    #     crit = crit.transpose()
    #     lines.set_ydata(resulty)
    #     markers.set_ydata(crit[1])
    #     markers.set_xdata(crit[0]*1e9)
    #     fig.canvas.get_tk_widget().update() # process events
    #     
    #     draw()
    # for f in np.linspace(1,0.01,100):
    #     resultx, resulty, crit=critical_point_interpolate(data, critical_points, f,nm)
    #     crit = crit.transpose()
    #     
    #     lines.set_ydata(resulty)
    #     markers.set_ydata(crit[1])
    #     markers.set_xdata(crit[0]*1e9)
    #     
    #     fig.canvas.get_tk_widget().update() # process events
    #     
    #     draw()
    #     
    # ioff()
    # show()
    # # legend()
    #



import collections, os, platform


def flatten(l):
    for el in l:
        if isinstance(el, collections.Iterable) and not isinstance(el, str):
            yield from flatten(el)
        else:
            yield el


def open_with_os(path):
    sys = platform.system()

    if sys == "Darwin":
        os.system('open "%s"' % path)
    elif sys == "Windows":
        os.startfile(path)
    elif sys == "Linux":
        os.system('evince "%s"' % path)
    else:
        raise NotImplementedError("Unable to open files in this particular system: %s" % sys)



import numpy as np
import matplotlib.pyplot as plt
import codecs
from solcore.absorption_calculator import OptiStack
from solcore.absorption_calculator import calculate_ellipsometry
from scipy.optimize import minimize
from typing import Optional, List, Dict, Collection


class EllipsometryData:
    """ This class is used to open and arrange in a sensible way ellipsometry datafiles created with the WVASE
    software."""

    def __init__(self, path: str) -> None:
        """ Creates an ellipsometry data object and loads data into it the ellipsometry data from a file.

        :param path: Path to the file
        """

        self.comment: str = ''
        self.meas_setup: str = ''
        self.units: str = ''
        self.angles: List[float] = []

        self.data: Dict = {}
        self.dpol: Dict = {}
        self._load_data(path)

    def _load_data(self, path: str) -> None:
        """ This is the function that actually loads the data

        :param path: Path to the file
        :return: None
        """

        with codecs.open(path, "r", encoding='utf-8', errors='ignore') as f:
            lines = f.readlines()
            self.comment = lines[0]
            self.meas_setup = lines[1]

            for l in lines[2:]:
                data = l.strip().split()

                if data[0] in ['nm', '1/cm']:
                    self.units = data[0]
                    continue
                elif data[0][0] in '0123456789':
                    s = 0
                    key = float(data[1])
                elif data[0] in ['E', 'dpolE']:
                    s = 1
                    key = float(data[2])
                else:
                    continue

                if key not in self.data.keys():
                    self.angles.append(key)
                    self.data[key] = []
                    self.dpol[key] = []

                if data[0] == 'E' or data[0][0] in '0123456789':
                    self.data[key].append([])
                    self.data[key][-1].append(float(data[0 + s]))  # wl
                    self.data[key][-1].append(float(data[2 + s]))  # Psi
                    self.data[key][-1].append(float(data[4 + s]))  # err Psi
                    self.data[key][-1].append(float(data[3 + s]))  # Delta
                    self.data[key][-1].append(float(data[5 + s]))  # err Delta

                elif data[0] == 'dpolE':
                    self.dpol[key].append([])
                    self.dpol[key][-1].append(float(data[1]))  # wl
                    self.dpol[key][-1].append(float(data[3]))  # dpol
                    self.dpol[key][-1].append(float(data[4]))  # err dpol

        for key in self.data.keys():
            self.data[key] = np.array(self.data[key]).T

            if 'cm' in self.units:
                self.data[key][0] = 10000. / self.data[key][0]
            else:
                self.data[key][0] /= 1000

        self.units = 'um'

    def append_data(self, path: str) -> None:
        """ Ellipsometry data from more than one source can be combined in a single object. This function does that.

        :param path: Path to the file with data we want to add to the current object.
        :return:
        """

        new_data = EllipsometryData(path)

        for key in new_data.data.keys():
            if key not in self.data.keys():
                self.data[key] = new_data.data[key]
                self.dpol[key] = new_data.dpol[key]
            else:
                self.data[key] = np.hstack((new_data.data[key], self.data[key]))
                self.data[key] = self.data[key][:, np.argsort(self.data[key][0], axis=0)]

                try:
                    self.dpol[key] = np.vstack((new_data.dpol[key], self.dpol[key]))
                    self.dpol[key] = self.dpol[key][:, np.argsort(self.dpol[key][0], axis=0)]
                except:
                    pass

    def plot(self, log: bool = True):
        """ Plots the ellipsometry data contained in this object in a sensible way.

        :param log: True/False for ploting logarithmic X scale.
        :return: None
        """
        fig, ax = plt.subplots(nrows=2, ncols=1, sharex=True)

        if log:
            for key in self.data.keys():
                ax[0].semilogx(self.data[key][0], self.data[key][1], 'o', mfc='none')
                ax[1].semilogx(self.data[key][0], self.data[key][3], 'o', mfc='none')
        else:
            for key in self.data.keys():
                ax[0].plot(self.data[key][0], self.data[key][1], 'o', mfc='none')
                ax[1].plot(self.data[key][0], self.data[key][3], 'o', mfc='none')

        plt.xlabel('Wavelength (µm)')
        ax[0].set_ylabel(r'$\Psi$ (º)')
        ax[1].set_ylabel(r'$\Delta$ (º)')
        plt.show()


class EllipsometryFitter:
    """ Class that contains all the tools to fit the properties of a OptiStack model to ellipsometry data."""

    def __init__(self, data: EllipsometryData, stack: Optional[OptiStack] = None,
                 vars: Optional[Collection] = None, wavelength_range: Optional[np.ndarray] = None) -> None:

        self.data = data
        self.stack = stack
        self.vars = vars
        self.range = wavelength_range

    def set_variables(self, vars: Collection) -> None:
        """ Sets the variables that will be used during the fitting. The input variable "vars" is a list with as many
        elements as layers in the model. Each of these elemenets is, in turn, a list with the names of the free
        variables of that layer. If the layers are defined by a dielectric model - rather than experimental n and k data
        the variable names other than the thickness and e_inf must be inside a list.

        vars =  [   [],
                    ['width', ['A1', 'A2']],
                    [['A1', 'E1'], ['A1', 'Br1']]
                ]

        The above variable list indicate that there are no free variables in the first layer, that the width and
        two amplitudes of the first model are free variables in the second layer, and for the first layer, that
        the free variables are the amplitude and energy of the first model, and the amplitude and broadening of the
        second one.

        The models are read in order. If there are free variables in the second model but not in the first one,
        then an empty list must be used. The last layer in the above example will look, in this case:

        [[], ['A1', 'Br1']

        :param vars: The list of free variables.
        :return: None.
        """
        assert len(vars) == len(self.stack.models), 'Error: the lenght of the vars list must be equal to the number ' \
                                                    'of layers in the stack = {}'.format(self.stack.num_layers)
        self.vars = vars

    def update_variables(self, v: List[float]) -> None:
        """ Updates the stack with the values of the variables contained in v. The meaning of each of the elements in v
        is given by the definitions in self.vars.

        :param v: An array with the new values of the variables.
        :return: None
        """
        k = 0
        for i, layer in enumerate(self.vars):
            # Loop over the layers
            if len(layer) > 0:
                model = 0
                for j, param in enumerate(layer):
                    # Loop over the parameterts of the layer
                    if param == 'width':
                        self.stack.widths[i] = v[k]
                        k += 1
                    # Parameters of the DielectricModel of the layer
                    elif param == 'e_inf':
                        self.stack.models[i].e_inf = v[k]
                        k += 1
                    elif param == 'e_gap':
                        self.stack.models[i].e_gap = v[k]
                        k += 1
                    else:
                        for val in param:
                            # Loop over the parameters of each oscillator within the DielectricModel
                            self.stack.models[i][model].__dict__[val] = v[k]
                            k += 1
                        model += 1

    def get_variables(self) -> List[float]:
        """ Provides a list with the current values of the free variables.

        :return: The values of the free variables, in the order they appear in self.vars.
        """
        out = []

        for i, layer in enumerate(self.vars):
            # Loop over the layers
            if len(layer) > 0:
                model = 0
                for j, param in enumerate(layer):
                    # Loop over the parameterts of the layer
                    if param == 'width':
                        out.append(self.stack.widths[i])
                    # Parameters of the DielectricModel of the layer
                    elif param == 'e_inf':
                        out.append(self.stack.models[i].e_inf)
                    elif param == 'e_gap':
                        out.append(self.stack.models[i].e_gap)
                    else:
                        for val in param:
                            # Loop over the parameters of each oscillator within the DielectricModel
                            out.append(self.stack.models[i][model].__dict__[val])
                        model += 1

        return out

    def set_range(self, wavelength_range: List[float]) -> None:
        """ Set the wavelengths in which to make the ellipsometry calculations. Ideally, it should be dense enought to
        capture all the features of the dielectric functions but sparse enought so the calculations are fast.

        :param wavelength_range: An array with the wavelength range in nanometers.
        :return: None
        """

        self.range = wavelength_range

    def mse(self, v: List[float] = None) -> np.ndarray:
        """ Calculates the mean-squared error between the experimental data and the modelled one. If the vars input
        is given, the variables are first updated and then the MSE is calculated.

        Since the experimental data might have a lot of wavelength points, we use a reduced number of them as defined
        in self.range in order to speed up the calculation of the ellipsometric parameters. The modelled data is then
        interpolated to the experimental one in order to calculate the MSE. While this will lead to higher uncertainties
        when the experimental data varies rapidly, it makes the calculation much faster.

        :param v: The free variables to update.
        :return: The mean-squared error between the experimental data and the modelled one
        """
        assert self.range is not None, 'Error: A wavelength range must be defined before calculating the MSE'

        if v is not None:
            self.update_variables(v)
            num_var = len(v)
        else:
            num_var = 0

        out = 0.
        elements = 0.

        mod_data = calculate_ellipsometry(self.stack, self.range, angle=self.data.angles)

        for i, ang in enumerate(self.data.angles):
            new_wl = self.data.data[ang][0] * 1000.
            elements += len(new_wl)
            psi = np.interp(new_wl, self.range, mod_data['psi'][:, i])
            delta = np.interp(new_wl, self.range, mod_data['Delta'][:, i])

            p = (self.data.data[ang][1] - psi) / self.data.data[ang][1]
            d = (self.data.data[ang][3] - delta) / self.data.data[ang][4]

            out += np.sum(p ** 2 + d ** 2)

        out = np.sqrt(1. / (2 * elements - num_var) * out)

        return np.array(out)

    def fit(self, show: bool = True) -> None:
        """ Calculate the best fit of the model to the data by minimising the mean-squared error using as free varaibles
         those defined in self.vars. The hessian matrix is used to calculate the 90% confident limits and the
         correlation matrix.

        :param show: If the result of the fit must be printed.
        :return:
        """

        guess = np.array(self.get_variables())

        def fun(x):
            return self.mse(x)

        self.fit_information = minimize(fun, guess, method='BFGS', options={'gtol': 1e-02, 'disp': True})
        self.calculate_uncertainties()
        self.update_variables(self.fit_information['x'])

        if show:
            print(self.fit_information)

    def calculate_uncertainties(self) -> None:
        """ Calculates the uncertainties in the calculated optimum parameters. For that, it uses the inverse of the
         hessian (or curvature matrix) C provided by the fitting algorithm. Two quantities are defined:

        Error for parameter i (= 90% confidence limit):
        .. math:: Err_{ii} = 1.65\\sqrt{MSE*C_{ii} }

        Correlation matrix:
        .. math:: CM_{ij} = \\frac{C_{ij}} {\\sqrt{C_{ii}} \\sqrt{C_{jj}}}

        The diagonals of the CM are always equal to one and the off-diagonal terms represent how correlated are the
        corresponding variables. Highly correlated variables (absolute values close to 1) mean that the minimum of the
        MSE is not very sharp with respect to changes between those two variables.

        The above information is stored in self.errors and self.correlations, respectively.

        :return: None
        """

        C = self.fit_information['hess_inv']
        MSE = self.fit_information['fun']

        self.errors = 1.65 * np.sqrt(MSE * np.diag(C))
        self.correlations = np.zeros_like(C)

        for i in range(len(C[0])):
            for j in range(len(C[0])):
                self.correlations[i, j] = C[i, j] / (np.sqrt(C[i, i]) * np.sqrt(C[j, j]))

    def plot(self, log: bool = True) -> None:
        """ Plots the expeirmental and the modelled data.

        :param log: If the X scale should be log.
        :return: None
        """

        mod_data = calculate_ellipsometry(self.stack, self.range, angle=self.data.angles)

        fig, ax = plt.subplots(nrows=2, ncols=1, sharex=True)
        min_wl = 0
        max_wl = 1e6

        if log:
            for i, ang in enumerate(self.data.angles):
                new_wl = self.data.data[ang][0]
                if min(new_wl) > min_wl: min_wl = min(new_wl)
                if max(new_wl) < max_wl: max_wl = max(new_wl)

                psi = np.interp(new_wl, self.range / 1000., mod_data['psi'][:, i])
                delta = np.interp(new_wl, self.range / 1000., mod_data['Delta'][:, i])

                ax[0].semilogx(new_wl, self.data.data[ang][1], 'o', mfc='none')
                ax[1].semilogx(new_wl, self.data.data[ang][3], 'o', mfc='none')
                ax[0].semilogx(new_wl, psi, 'r')
                ax[1].semilogx(new_wl, delta, 'b')

        else:
            for i, ang in enumerate(self.data.angles):
                new_wl = self.data.data[ang][0] * 1000.
                if min(new_wl) > min_wl: min_wl = min(new_wl)
                if max(new_wl) < max_wl: max_wl = max(new_wl)

                psi = np.interp(new_wl, self.range / 1000., mod_data['psi'][:, i])
                delta = np.interp(new_wl, self.range / 1000., mod_data['Delta'][:, i])

                ax[0].plot(new_wl, self.data.data[ang][1], 'o', mfc='none')
                ax[1].plot(new_wl, self.data.data[ang][3], 'o', mfc='none')
                ax[0].semilogx(new_wl, psi, 'r')
                ax[1].semilogx(new_wl, delta, 'b')

        ax[0].set_xlim((min_wl, max_wl))
        ax[1].set_xlim((min_wl, max_wl))

        x = [0.3, 0.4, 0.5, 0.7, 1, 2, 3, 4, 5, 7, 10, 20]
        plt.xticks(x, x)
        plt.xlabel('Wavelength (µm)')
        ax[0].set_ylabel(r'$\Psi$ (º)')
        ax[1].set_ylabel(r'$\Delta$ (º)')
        plt.show()

    def __repr__(self) -> str:

        out = ''
        for i, layer in enumerate(self.stack.models):
            out += '\n LAYER {}:\n'.format(i)
            out += 'Width:\t {} nm\n'.format(self.stack.widths[i])
            out += layer.__repr__()
            out += '\n'

        return out


if __name__ == '__main__':
    from solcore.absorption_calculator.dielectric_constant_models import Poles, Lorentz, Drude, Gauss, Cauchy
    from solcore.absorption_calculator import DielectricConstantModel

    import matplotlib.pyplot as plt

    # Experimental data
    filename = '/Users/diego/Dropbox/WorkIC/Data/Imperial/RTA_samples/Ellipsometry_IR/nSi_bare.dat'
    filename2 = '/Users/diego/Dropbox/WorkIC/Data/Imperial/RTA_samples/Ellipsometry_VIS/160920/ellipvis_nsi2.dat'
    data = EllipsometryData(filename)
    data.append_data(filename2)
    # data.plot()

    # Models
    silicon_filename = '/Users/diego/Dropbox/WorkIC/Data/Materials_database/FTIR_mat/silicon.mat'
    si_data = np.loadtxt(silicon_filename, skiprows=3, unpack=True)
    si_data[0] = 1.24 / si_data[0]
    n = np.sqrt(si_data[1] + si_data[2] * 1.j)
    si_data[1] = np.real(n)
    si_data[2] = np.imag(n)

    # model_substrate = [300, si_data[0]*1000, si_data[1], si_data[2]]
    #

    drud = Drude(An=2.26598948, Brn=0.049)
    model_substrate = [300, si_data[0] * 1000, si_data[1], si_data[2],
                       DielectricConstantModel(e_inf=11.71255715, oscillators=[drud]), [2000, 500, 0]]

    siO2_filename = '/Users/diego/Dropbox/WorkIC/Data/Materials_database/FTIR_mat/SiO2_thermal.mat'
    sio2_data = np.loadtxt(siO2_filename, skiprows=3, unpack=True)

    sio2_data[0] = 1.24 / sio2_data[0]
    n = np.sqrt(sio2_data[1] + sio2_data[2] * 1.j)
    sio2_data[1] = np.real(n)
    sio2_data[2] = np.imag(n)
    model_sio2 = [1.5, sio2_data[0] * 1000, sio2_data[1], sio2_data[2]]

    # gau1 = Gauss(A=2.8468, Ec=0.1299, Br=0.0111)
    # gau2 = Gauss(A=7.192, Ec=0.058331, Br=0.01682)
    # gau3 = Gauss(A=1.9737, Ec=0.13991, Br=0.02144)
    # cau = Cauchy(An=1.5, Bn=0.002, Ak=1, Bk=0.2580645161)

    # model_sio2 = [1.5, DielectricConstantModel(e_inf=3, oscillators=[cau])]

    stack = OptiStack([model_sio2, model_substrate])

    # Fitting stuff
    fit = EllipsometryFitter(data, stack)
    variables = [['width'], ['e_inf', ['An']]]
    fit.set_variables(variables)
    wl = np.logspace(np.log10(300), np.log10(20000), 200)
    fit.set_range(wl)

    fit.plot()

    # fit.fit()
    #
    # print(fit.errors)
    # print(fit.correlations)

    # Ploting the result
    # fit.plot()

    # fit.get_all_vars()
    #
    # n_fin = fit.stack.models[0].n_and_k(wl)
    #
    # plt.semilogx(wl, np.real(n_ini))
    # plt.semilogx(wl, np.real(n_fin))
    # plt.show()



# from: https://github.com/HugoGuillen/refractiveindex.info-sqlite

import os
import yaml
import sys
import argparse
import numpy
import scipy.interpolate

class DBMaterial:
    """ Material class"""
    def __init__(self, filename,interpolation_points=100,empty=False):
        """

        :param filename:
        """
        self.refractiveIndex = None
        self.extinctionCoefficient = None
        self.points = interpolation_points
        if empty:
            return

        f = open(filename)
        try:
            material = yaml.safe_load(f)
        except:
            raise Exception('Bad Material YAML File.')
        finally:
            f.close()

        previous_formula = False
        for data in material['DATA']:
            if (data['type'].split())[0] == 'tabulated':
                rows = data['data'].split('\n')
                splitrows = [c.split() for c in rows]
                wavelengths = []
                n = []
                k = []
                for s in splitrows:
                    if len(s) > 0:
                        wavelengths.append(float(s[0]))
                        n.append(float(s[1]))
                        if len(s) > 2:
                            k.append(float(s[2]))
                self.points = len(wavelengths)

                if (data['type'].split())[1] == 'n':

                    if self.refractiveIndex is not None:
                        Exception('Bad Material YAML File')

                    self.refractiveIndex = RefractiveIndexData.SetupRefractiveIndex(formula=-1,
                                                                                    wavelengths=wavelengths,
                                                                                    values=n)
                elif (data['type'].split())[1] == 'k':

                    self.extinctionCoefficient = ExtinctionCoefficientData.SetupExtinctionCoefficient(wavelengths, n)

                    if previous_formula:
                        self.refractiveIndex = RefractiveIndexData.SetupRefractiveIndex(formula=formula,
                                                                                        rangeMin=rangeMin,
                                                                                        rangeMax=rangeMax,
                                                                                        coefficients=coefficients,
                                                                                        interpolation_points=self.points)
                elif (data['type'].split())[1] == 'nk':

                    if self.refractiveIndex is not None:
                        Exception('Bad Material YAML File')

                    self.refractiveIndex = RefractiveIndexData.SetupRefractiveIndex(formula=-1,
                                                                                    wavelengths=wavelengths,
                                                                                    values=n)
                    self.extinctionCoefficient = ExtinctionCoefficientData.SetupExtinctionCoefficient(wavelengths, k)
            elif (data['type'].split())[0] == 'formula':

                if self.refractiveIndex is not None:
                    Exception('Bad Material YAML File')

                formula = int((data['type'].split())[1])
                coefficients = [float(s) for s in data['coefficients'].split()]

                rangeMin = float(data['wavelength_range'].split()[0])
                rangeMax = float(data['wavelength_range'].split()[1])

                previous_formula = True
                self.refractiveIndex = RefractiveIndexData.SetupRefractiveIndex(formula=formula,
                                                                                rangeMin=rangeMin,
                                                                                rangeMax=rangeMax,
                                                                                coefficients=coefficients,
                                                                                interpolation_points=self.points)
        if self.refractiveIndex is not None:
            self.rangeMin = self.refractiveIndex.rangeMin
            self.rangeMax = self.refractiveIndex.rangeMax
        else:
            self.rangeMin = self.extinctionCoefficient.rangeMin
            self.rangeMax = self.extinctionCoefficient.rangeMax

    def get_refractiveindex(self, wavelength):
        """

        :param wavelength:
        :return: :raise Exception:
        """
        if self.refractiveIndex is None:
            raise Exception('No refractive index specified for this material')
        else:
            return self.refractiveIndex.get_refractiveindex(wavelength)

    def get_extinctioncoefficient(self, wavelength):
        """

        :param wavelength:
        :return: :raise NoExtinctionCoefficient:
        """
        if self.extinctionCoefficient is None:
            raise NoExtinctionCoefficient('No extinction coefficient specified for this material')
        else:
            return self.extinctionCoefficient.get_extinction_coefficient(wavelength)

    def get_complete_extinction(self):
        if self.has_extinction():
            return self.extinctionCoefficient.get_complete_extinction()
        else:
            return None

    def get_complete_refractive(self):
        if self.has_refractive():
            return self.refractiveIndex.get_complete_refractive()
        else:
            return None

    def has_refractive(self):
        return self.refractiveIndex is not None

    def has_extinction(self):
        return self.extinctionCoefficient is not None

    def get_page_info(self):
        return self.pageinfo

    def to_csv(self, output):
        refr = self.get_complete_refractive()
        ext = self.get_complete_extinction()
        #FizzFuzz
        if self.has_refractive() and self.has_extinction() and len(refr) == len(ext):
            header = "wl,n,k\n"
            output_f = open(output.replace(".csv","(nk).csv"),'w')
            output_f.write(header)
            for i in range(len(refr)):
                output_f.write(",".join(list(map(str,[refr[i][0],refr[i][1],ext[i][1]])))+"\n")
            output_f.close()
            print("Wrote",output.replace(".csv","(nk).csv"))
        else:
            if self.has_refractive():
                output_f = open(output.replace(".csv","(n).csv"),'w')
                header = "wl,n\n"
                output_f.write(header)
                for i in range(len(refr)):
                    output_f.write(",".join(list(map(str,[refr[i][0],refr[i][1]])))+"\n")
                output_f.close()
                print("Wrote",output.replace(".csv","(n).csv"))
            if self.has_extinction():
                output_f = open(output.replace(".csv","(k).csv"),'w')
                header = "wl,k\n"
                output_f.write(header)
                for i in range(len(ext)):
                    output_f.write(",".join(list(map(str,[ext[i][0],ext[i][1]])))+"\n")
                output_f.close()
                print("Wrote",output.replace(".csv","(k).csv"))

    def to_txt(self, output):
        refr = self.get_complete_refractive()
        ext = self.get_complete_extinction()
        #FizzFuzz

        if self.has_refractive():
            output_f = open(output.replace(".txt","_n.txt"),'w')
            for i in range(len(refr)):
                output_f.write(" ".join(list(map(str,[refr[i][0]*1e-6,refr[i][1]])))+"\n")
            output_f.close()
            print("Wrote",output.replace(".txt","_n.txt"))
        else:
            print("No n data available, file not written")

        if self.has_extinction():
            output_f = open(output.replace(".txt","_k.txt"),'w')
            for i in range(len(ext)):
                output_f.write(" ".join(list(map(str,[ext[i][0]*1e-6,ext[i][1]])))+"\n")
            output_f.close()
            print("Wrote",output.replace(".txt","_k.txt"))
        else:
            print("No k data available, file not written")

    @staticmethod
    def FromLists(pageinfo,wavelengths_r=None,refractive=None,wavelengths_e=None,extinction=None):
        mat = DBMaterial("",empty=True)
        mat.pageinfo = pageinfo
        if refractive is not None:
            mat.refractiveIndex = TabulatedRefractiveIndexData.FromLists(wavelengths_r,refractive)
            mat.rangeMin = mat.refractiveIndex.rangeMin
            mat.rangeMax = mat.refractiveIndex.rangeMax
        if extinction is not None:
            mat.extinctionCoefficient = ExtinctionCoefficientData.FromLists(wavelengths_e,extinction)
            mat.rangeMin = mat.extinctionCoefficient.rangeMin
            mat.rangeMax = mat.extinctionCoefficient.rangeMax
        return mat
#
# Refractive Index
#
class RefractiveIndexData:
    """Abstract RefractiveIndex class"""

    @staticmethod
    def SetupRefractiveIndex(formula, **kwargs):
        """

        :param formula:
        :param kwargs:
        :return: :raise Exception:
        """
        if formula >= 0:
            return FormulaRefractiveIndexData(formula, **kwargs)
        elif formula == -1:
            return TabulatedRefractiveIndexData(**kwargs)
        else:
            raise Exception('Bad RefractiveIndex data type')

    def get_refractiveindex(self, wavelength):
        """

        :param wavelength:
        :raise NotImplementedError:
        """
        raise NotImplementedError('Different for functionally and experimentally defined materials')


class FormulaRefractiveIndexData:
    """Formula RefractiveIndex class"""

    def __init__(self, formula, rangeMin, rangeMax, coefficients,interpolation_points):
        """

        :param formula:
        :param rangeMin:
        :param rangeMax:
        :param coefficients:
        """
        self.formula = formula
        self.rangeMin = rangeMin
        self.rangeMax = rangeMax
        self.coefficients = coefficients
        self.interpolation_points = interpolation_points


    def get_complete_refractive(self):
        #print(self.rangeMin, self.rangeMax)
        wavelength = numpy.linspace(self.rangeMin, self.rangeMax,num=self.interpolation_points)
        extlist = [[wavelength[i], self.get_refractiveindex(wavelength[i] * 1000)] for i in range(len(wavelength))]
        #return numpy.array(extlist)
        return extlist

    def get_refractiveindex(self, wavelength):
        """

        :param wavelength:
        :return: :raise Exception:
        """
        wavelength /= 1000.0
        if self.rangeMin <= wavelength <= self.rangeMax:
            formula_type = self.formula
            coefficients = self.coefficients
            n = 0
            if formula_type == 1:  # Sellmeier
                nsq = 1 + coefficients[0]
                g = lambda c1, c2, w: c1 * (w ** 2) / (w ** 2 - c2 ** 2)
                for i in range(1, len(coefficients), 2):
                    nsq += g(coefficients[i], coefficients[i + 1], wavelength)
                n = numpy.sqrt(nsq)
            elif formula_type == 2:  # Sellmeier-2
                nsq = 1 + coefficients[0]
                g = lambda c1, c2, w: c1 * (w ** 2) / (w ** 2 - c2)
                for i in range(1, len(coefficients), 2):
                    nsq += g(coefficients[i], coefficients[i + 1], wavelength)
                n = numpy.sqrt(nsq)
            elif formula_type == 3:  # Polynomal
                g = lambda c1, c2, w: c1 * w ** c2
                nsq = coefficients[0]
                for i in range(1, len(coefficients), 2):
                    nsq += g(coefficients[i], coefficients[i + 1], wavelength)
                n = numpy.sqrt(nsq)
            elif formula_type == 4:  # RefractiveIndex.INFO
                g = lambda wl, ci, cj, ck, cl: ci * wl**cj / (wl**2 - ck**cl)
                n = coefficients[0]
                n += g(wavelength, *coefficients[1:5])
                n += g(wavelength, *coefficients[5:9])
                for kk in range(len(coefficients[9:]) // 2):
                    n += coefficients[9+kk] * wavelength**coefficients[9+kk+1]

                n = numpy.sqrt(n)
            elif formula_type == 5:  # Cauchy
                g = lambda c1, c2, w: c1 * w ** c2
                n = coefficients[0]
                for i in range(1, len(coefficients), 2):
                    n += g(coefficients[i], coefficients[i + 1], wavelength)
            elif formula_type == 6:  # Gasses
                n = 1 + coefficients[0]
                g = lambda c1, c2, w: c1 / (c2 - w ** (-2))
                for i in range(1, len(coefficients), 2):
                    n += g(coefficients[i], coefficients[i + 1], wavelength)
            elif formula_type == 7:  # Herzberger
                n = coefficients[0]
                n += coefficients[1] / (wavelength**2 - 0.028)
                n += coefficients[2] * (1 / (wavelength**2 - 0.028))**2
                for i, cc in enumerate(coefficients[3:]):
                    n += cc * wavelength**(2*(i+1))
            elif formula_type == 8:  # Retro
                n = coefficients[0]
                n += coefficients[1] * wavelength**2 / (wavelength**2 - coefficients[2])
                n += coefficients[3] * wavelength**2
                n = numpy.sqrt(-(2 * n + 1) / (n - 1))
            elif formula_type == 9:  # Exotic
                n = coefficients[0]
                n += coefficients[1] / (wavelength**2 - coefficients[2])
                n += coefficients[3] * (wavelength - coefficients[4]) / \
                     ((wavelength - coefficients[4])**2 + coefficients[5])
                n = numpy.sqrt(n)
            else:
                raise Exception('Bad formula type')

            return n
        else:
            raise Exception(
                'Wavelength {} is out of bounds. Correct range(um): ({}, {})'.format(wavelength, self.rangeMin,
                                                                                     self.rangeMax))


class TabulatedRefractiveIndexData:
    """Tabulated RefractiveIndex class"""

    def __init__(self, wavelengths, values):
        """

        :param wavelengths:
        :param values:
        """
        self.rangeMin = numpy.min(wavelengths)
        self.rangeMax = numpy.max(wavelengths)

        if self.rangeMin == self.rangeMax:
            self.refractiveFunction = values[0]
        else:
            self.refractiveFunction = scipy.interpolate.interp1d(wavelengths, values)

        self.wavelengths = wavelengths
        self.coefficients = values

    @staticmethod
    def FromLists(wavelengths,values):
        return TabulatedRefractiveIndexData(wavelengths,values)

    def get_refractiveindex(self, wavelength):
        """

        :param wavelength:
        :return: :raise Exception:
        """
        wavelength /= 1000.0
        if self.rangeMin == self.rangeMax and self.rangeMin == wavelength:
            return self.refractiveFunction
        elif self.rangeMin <= wavelength <= self.rangeMax and self.rangeMin != self.rangeMax:
            return self.refractiveFunction(wavelength)
        else:
            raise Exception(
                'Wavelength {} is out of bounds. Correct range(um): ({}, {})'.format(wavelength, self.rangeMin,
                                                                                     self.rangeMax))
    def get_complete_refractive(self):
        extlist =  [[self.wavelengths[i],self.coefficients[i]] for i in range(len(self.wavelengths))]
        #return numpy.array(extlist)
        return extlist


#
# Extinction Coefficient
#
class ExtinctionCoefficientData:
    """ExtinctionCofficient class"""

    @staticmethod
    def SetupExtinctionCoefficient(wavelengths, values):
        """

        :param wavelengths:
        :param values:
        :return:
        """
        return ExtinctionCoefficientData(wavelengths, values)

    @staticmethod
    def FromLists(wavelengths,values):
        return ExtinctionCoefficientData(wavelengths, values)

    def __init__(self, wavelengths, coefficients):
        """

        :param wavelengths:
        :param coefficients:
        """
        self.extCoeffFunction = scipy.interpolate.interp1d(wavelengths, coefficients)
        self.rangeMin = numpy.min(wavelengths)
        self.rangeMax = numpy.max(wavelengths)
        self.wavelengths = wavelengths
        self.coefficients = coefficients

    def get_extinction_coefficient(self, wavelength):
        """

        :param wavelength:
        :return: :raise Exception:
        """
        wavelength /= 1000.0
        if self.rangeMin <= wavelength <= self.rangeMax:
            return self.extCoeffFunction(wavelength)
        else:
            raise Exception(
                'Wavelength {} is out of bounds. Correct range(um): ({}, {})'.format(wavelength, self.rangeMin,
                                                                                     self.rangeMax))
    def get_complete_extinction(self):
        extlist =  [[self.wavelengths[i],self.coefficients[i]] for i in range(len(self.wavelengths))]
        #return numpy.array(extlist)
        return extlist

#
# Custom Exceptions
#
class FormulaNotImplemented(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class NoExtinctionCoefficient(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)



from pytest import approx


def test_register_conversion_function():
    from solcore.light_source.light_source import (
        REGISTERED_CONVERTERS,
        power_density_per_nm,
    )

    name = power_density_per_nm.__name__
    assert name in REGISTERED_CONVERTERS
    assert REGISTERED_CONVERTERS[name] == power_density_per_nm


def test_power_density_per_nm(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import power_density_per_nm

    expected, sp_fun = gauss_spectrum
    actual = power_density_per_nm(sp_fun, wavelength)
    assert actual == approx(expected)


def test_photon_flux_per_nm(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import photon_flux_per_nm
    from solcore.constants import c, h

    sp, sp_fun = gauss_spectrum
    expected = sp / (c * h * 1e9 / wavelength)

    actual = photon_flux_per_nm(sp_fun, wavelength)
    assert actual == approx(expected)


def test_power_density_per_m(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import power_density_per_m
    import numpy as np

    sp, sp_fun = gauss_spectrum
    expected = np.trapz(sp, wavelength)

    wl_m = wavelength * 1.0e-9
    actual = power_density_per_m(sp_fun, wl_m)
    actual = np.trapz(actual, wl_m)
    assert actual == approx(expected)


def test_photon_flux_per_m(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import photon_flux_per_m
    from solcore.constants import c, h

    sp, sp_fun = gauss_spectrum

    wl_m = wavelength * 1e-9
    expected = sp / (c * h / wavelength)
    actual = photon_flux_per_m(sp_fun, wl_m)
    assert actual == approx(expected)


def test_power_density_per_ev(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import power_density_per_ev
    from solcore import spectral_conversion_nm_ev

    sp, sp_fun = gauss_spectrum
    ev, expected = spectral_conversion_nm_ev(wavelength, sp)

    actual = power_density_per_ev(sp_fun, ev)
    assert actual == approx(expected)


def test_photon_flux_per_ev(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import photon_flux_per_ev
    from solcore import spectral_conversion_nm_ev
    from solcore.constants import q

    sp, sp_fun = gauss_spectrum
    ev, expected = spectral_conversion_nm_ev(wavelength, sp)

    actual = photon_flux_per_ev(sp_fun, ev)
    assert actual * q * ev == approx(expected)


def test_power_density_per_joule(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import power_density_per_joule
    from solcore import spectral_conversion_nm_ev
    from solcore.constants import q

    sp, sp_fun = gauss_spectrum
    ev, expected = spectral_conversion_nm_ev(wavelength, sp)

    actual = power_density_per_joule(sp_fun, ev * q)
    assert actual * q == approx(expected)


def test_photon_flux_per_joule(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import photon_flux_per_joule
    from solcore import spectral_conversion_nm_ev
    from solcore.constants import q

    sp, sp_fun = gauss_spectrum
    ev, expected = spectral_conversion_nm_ev(wavelength, sp)

    actual = photon_flux_per_joule(sp_fun, ev * q)
    assert actual * q ** 2 * ev == approx(expected)


def test_power_density_per_hz(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import power_density_per_hz
    from solcore import spectral_conversion_nm_hz

    sp, sp_fun = gauss_spectrum
    hz, expected = spectral_conversion_nm_hz(wavelength, sp)

    actual = power_density_per_hz(sp_fun, hz)
    assert actual == approx(expected)


def test_photon_flux_per_hz(wavelength, gauss_spectrum):
    from solcore.light_source.light_source import photon_flux_per_hz
    from solcore import spectral_conversion_nm_hz
    from solcore.constants import h

    sp, sp_fun = gauss_spectrum
    hz, expected = spectral_conversion_nm_hz(wavelength, sp)

    actual = photon_flux_per_hz(sp_fun, hz)
    assert actual * h * hz == approx(expected)

def test_wavelength_array_consistency():
    #GH 147
    from solcore.light_source import LightSource
    import numpy as np

    wl_arr = np.linspace(300, 1200, 10)*1e-9
    wl_arr2 = np.linspace(300, 1200, 20)*1e-9

    ls = LightSource(source_type='standard', version='AM1.5g', x=wl_arr,
                             output_units='photon_flux_per_m')
    assert ls.spectrum()[1].shape == wl_arr.shape
    assert ls.spectrum(x=wl_arr2)[1].shape == wl_arr2.shape



from solcore import verbose
from typing import Any, Dict, Callable


class Singleton(type):  # META!
    """ A metaclass that restricts children to one instance.
        Attempting to create a second instance simply returns the first instance.
    """
    # All the Singleton instances live here...
    _dufftown: Dict['Singleton', 'Singleton'] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> 'Singleton':
        if cls not in cls._dufftown:  # ... or if they don't yet, it's about time!
            # print("Instantiating cls", cls)
            cls._dufftown[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._dufftown[cls]

    def breakout(method: 'Singleton') -> 'Singleton':
        """ Decorator for singleton methods that should be available in global scope
        """
        method.break_me_out = None  # or, you know, anything.
        return method

    @staticmethod
    class Redirect_to_singleton_method:
        """ Class that wraps the method that has been defined as 'breakout' with its own class, so it can be called as a
            standalone function.
        """

        def __init__(self, cls: 'Singleton', function: Callable) -> None:
            self.singleton_class = cls  # LABEL COME_BACK_HERE; here instead.;RETURN;
            self.function = function

        def __call__(self, *args: Any, **kwargs: Any) -> str:
            """  Call the singleton method and return the result """
            return self.function(self.singleton_class(), *args, **kwargs)

    class breakoutFunctions:
        """ Decorator for singletons which should make some of their methods available in a certain namespace.
            This is identical to the 'breakoutClass' function defined in 'source_managed_class', but has the option of
            making the function available not in the global scope but somewhere else. """

        def __init__(self, namespace: str) -> None:
            self.namespace_to_breakout_into = namespace

        def __call__(self, cls: 'Singleton') -> 'Singleton':
            """ And this is really the decorator for singletons which should make some of their methods available
            somewhere else
            """
            if verbose:
                print("Making functions from '{}' available "
                      "in the '{}' scope: ".format(cls.__name__, self.namespace_to_breakout_into.__name__), end="")

            # We scan all the methods of the class
            for funcname in (cls.__dict__):
                func = getattr(cls, funcname)

                # And choose those tagged with the 'breakout' decorator
                if hasattr(func, "break_me_out"):
                    # We remove the decorator-ralated atribute...
                    del func.break_me_out
                    # ... and make the function 'func' available in the given scope
                    setattr(self.namespace_to_breakout_into, funcname,
                            Singleton.Redirect_to_singleton_method(cls, func))

                    if verbose:
                        print(funcname, end=" ")

            if verbose:
                print()

            return cls


if __name__ == "__main__":
    class test:
        __name__ = 'ASC_examples'
        pass


    moo = test()


    @Singleton.breakoutFunctions(moo)
    class ForeverAlone(metaclass=Singleton):
        @Singleton.breakout
        def wait_you_havent_instantiated_anything(self) -> str:
            return "YES I DID YOU JUST MISSED IT: " + str(self)

        @Singleton.breakout
        def a_global_instance_method_surely_not(self) -> str:
            return "YES SURELY."

        @Singleton.breakout
        def rofl_this_will_never_work(self, end_symbol: str,
                                      end_number: int) -> str:
            return "WHO'S LAUGHING NOW" + end_symbol * end_number


    a = moo.wait_you_havent_instantiated_anything()
    b = moo.a_global_instance_method_surely_not()
    c = moo.rofl_this_will_never_work('!', end_number=3)

    print("moo.wait_you_havent_instantiated_anything() -> ", a)
    print("moo.a_global_instance_method_surely_not() -> ", b)
    print("moo.rofl_this_will_never_work('!', end_number=3) -> ", c)

    # should output:
    # wait_you_havent_instantiated_anything() ->  YES I DID YOU JUST MISSED IT: <__main__.ForeverAlone object at 0x10c377b90>
    # a_global_instance_method_surely_not() ->  YES SURELY.
    # rofl_this_will_never_work('!', end_number=3) ->  WHO'S LAUGHING NOW!!!



from pytest import fixture
import numpy as np
from unittest.mock import patch
import os
import tempfile


@fixture(autouse=True)
def solcore_user(tmp_path):
    os.environ["SOLCORE_USER_DATA"] = str(tmp_path)


def patch_plots(function):
    from functools import wraps

    @wraps(function)
    def decorated(*args, **kwargs):

        with patch("matplotlib.pyplot.show", lambda *x, **y: None):
            import matplotlib

            matplotlib.use("Agg")
            return function(*args, **kwargs)

    return decorated


@fixture
def wavelength():
    return np.linspace(300, 1200)


@fixture
def gauss_spectrum(wavelength):
    from scipy.interpolate import interp1d

    centre = wavelength.mean()
    width = (wavelength.max() - wavelength.min()) / 6
    sp = np.exp(-(wavelength - centre) ** 2 / width)
    return sp, interp1d(wavelength, sp, bounds_error=False, fill_value=0)


def da_light_source():
    from solcore.light_source import LightSource

    light_source = LightSource(source_type="standard", version="AM1.5g")
    return light_source


def da_options():
    from solcore.state import State

    options = State()
    wl = np.linspace(290, 700, 150) * 1e-9
    options.T = np.random.uniform(250, 350)
    options.wavelength = wl
    options.light_source = da_light_source()
    options.position = None
    options.internal_voltages = np.linspace(-6, 4, 20)

    return options


def junction(nd_top, na_top, nd_bottom, na_bottom):
    from solcore.structure import Junction, Layer
    from solcore import si, material
    from solcore.solar_cell import SolarCell
    from solcore.constants import vacuum_permittivity
    from solcore.solar_cell_solver import prepare_solar_cell
    from solcore.optics import solve_beer_lambert

    Lp = np.power(10, np.random.uniform(-8, -6))
    Ln = np.power(10, np.random.uniform(-8, -6))

    AlInP = material("AlInP")
    InGaP = material("GaInP")
    window_material = AlInP(Al=0.52)
    top_cell_n_material = InGaP(
        In=0.48,
        Nd=nd_top,
        Na=na_top,
        hole_diffusion_length=Lp,
        electron_diffusion_length=Ln,
    )
    top_cell_p_material = InGaP(
        In=0.48,
        Nd=nd_bottom,
        Na=na_bottom,
        hole_diffusion_length=Lp,
        electron_diffusion_length=Ln,
    )

    rel_perm = np.random.uniform(1, 20)
    for mat in [top_cell_n_material, top_cell_p_material]:
        mat.permittivity = rel_perm * vacuum_permittivity

    n_width = np.random.uniform(500, 1000) * 1e-9
    p_width = np.random.uniform(3000, 5000) * 1e-9

    test_junc = SolarCell(
        [
            Junction(
                [
                    Layer(si("25nm"), material=window_material, role="window"),
                    Layer(n_width, material=top_cell_n_material, role="emitter"),
                    Layer(p_width, material=top_cell_p_material, role="base"),
                ],
                sn=1,
                sp=1,
                kind="DA",
            )
        ]
    )

    options = da_options()
    options.light_source = da_light_source()
    prepare_solar_cell(test_junc, options)
    solve_beer_lambert(test_junc, options)

    return test_junc, options


@fixture
def np_junction():

    Nd_top = np.power(10, np.random.uniform(24, 26))
    Na_bottom = np.power(10, np.random.uniform(23, 25))

    Na_top = 0
    Nd_bottom = 0

    test_junc, options = junction(Nd_top, Na_top, Nd_bottom, Na_bottom)
    return test_junc, options


@fixture
def pn_junction():

    Na_top = np.power(10, np.random.uniform(24, 26))
    Nd_bottom = np.power(10, np.random.uniform(23, 25))

    Na_bottom = 0
    Nd_top = 0

    test_junc, options = junction(Nd_top, Na_top, Nd_bottom, Na_bottom)
    return test_junc, options


@fixture
def prepare_test_cell():

    from solcore import material, si
    from solcore.structure import Junction, Layer, TunnelJunction
    from solcore.solar_cell import SolarCell

    GaAs = material("GaAs")()
    MgF2 = material("MgF2")()
    TiO2 = material("TiO2")()
    Ge = material("Ge")()

    widths = np.random.rand(9) * 200

    solar_cell = SolarCell(
        [
            Layer(si(widths[0], "nm"), material=MgF2),
            Layer(si(widths[1], "nm"), material=TiO2),
            Junction(
                [
                    Layer(si(widths[2], "nm"), material=GaAs, role="window"),
                    Layer(si(widths[3], "nm"), material=GaAs, role="emitter"),
                    Layer(si(widths[4], "nm"), material=GaAs, role="base"),
                ],
                kind="DA",
            ),
            TunnelJunction(
                [
                    Layer(si(widths[5], "nm"), material=GaAs),
                    Layer(si(widths[6], "nm"), material=GaAs),
                ]
            ),
            Junction(
                [
                    Layer(si(widths[7], "nm"), material=Ge, role="emitter"),
                    Layer(si(widths[8], "nm"), material=Ge, role="base"),
                ],
                kind="PDD",
            ),
        ]
    )

    return solar_cell, widths


@fixture
def AlGaAs():
    from solcore import material
    from solcore.structure import Layer, Junction
    from solcore import si

    T = 298
    # We create the other materials we need for the device
    window = material("AlGaAs")(T=T, Na=5e24, Al=0.8)
    p_AlGaAs = material("AlGaAs")(T=T, Na=1e24, Al=0.4)
    n_AlGaAs = material("AlGaAs")(T=T, Nd=8e22, Al=0.4)
    bsf = material("AlGaAs")(T=T, Nd=2e24, Al=0.6)

    return Junction(
        [
            Layer(width=si("30nm"), material=window, role="Window"),
            Layer(width=si("150nm"), material=p_AlGaAs, role="Emitter"),
            Layer(width=si("1000nm"), material=n_AlGaAs, role="Base"),
            Layer(width=si("200nm"), material=bsf, role="BSF"),
        ],
        sn=1e6,
        sp=1e6,
        T=T,
        kind="PDD",
    )


@fixture
def light_source(wavelength):
    from solcore.light_source import LightSource

    return LightSource(
        source_type="standard",
        version="AM1.5g",
        x=np.linspace(350, 1000, 301) * 1e-9,
        output_units="photon_flux_per_m",
        concentration=1,
    )



from .spectral2 import get_default_spectral2_object, calculate_spectrum_spectral2
from .smarts import get_default_smarts_object, calculate_spectrum_smarts, SmartsSolverError
from .light_source import reference_spectra, LightSource






# From: https: // github.com / HugoGuillen / refractiveindex.info - sqlite

from collections import namedtuple,OrderedDict
import os
import yaml
import sqlite3
import numpy as np

# from solcore.material_data.refractiveindex_info_DB import dbmaterial
from solcore.material_data.refractiveindex_info_DB import DBMaterial

Shelf = namedtuple('Shelf', ['shelf', 'name'])
Book = namedtuple('Book', ['book', 'name'])
Page = namedtuple('Page', ['page', 'name', 'path'])
Entry = namedtuple('Entry',['id','shelf','book','page'])

# Check latest available database in https://refractiveindex.info/download.php
_riiurl = "https://refractiveindex.info/download/database/rii-database-2018-07-01.zip"

class Database:

    def __init__(self, sqlitedbpath):
        self.db_path = sqlitedbpath
        if not os.path.isfile(sqlitedbpath):
            print("Database file not found.")
        else:
            print("Database file found at",sqlitedbpath)

    def create_database_from_folder(self, yml_database_path, interpolation_points=100):
        create_sqlite_database(yml_database_path, self.db_path,interpolation_points=interpolation_points)

    def create_database_from_url(self,interpolation_points=100,riiurl=_riiurl, outputfolder=""):
        Database.DownloadRIIzip(riiurl=riiurl, outputfolder=outputfolder)
        outputfolder = os.path.join(outputfolder, "database")
        self.create_database_from_folder(outputfolder, interpolation_points=interpolation_points)
        pass

    def check_url_version(self):
        print(_riiurl)

    def search_custom(self,sqlquery):
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        c.execute(sqlquery)
        results = c.fetchall()
        if len(results)==0:
            print("No results found.")
        else:
            print(len(results),"results found.")
        conn.close()
        return results

    def search_pages(self,term="",exact=False):
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        if not exact:
            c.execute('SELECT * FROM pages WHERE shelf like ? or book like ? or page like ? or filepath like ?', ["%"+term+"%" for i in range(4)])
        else:
            c.execute('SELECT * FROM pages WHERE shelf like ? or book like ? or page like ? or filepath like ?', [term for i in range(4)])
        results = c.fetchall()
        if len(results)==0:
            print("No results found.")
        else:
            print(len(results),"results found.")
            columns = self._get_pages_columns()
            print("\t".join(columns))
            for r in results:
                print("\t".join(map(str,r[:])))
        conn.close()
        #return results

    def search_id(self,pageid):
        info = self._get_page_info(pageid)
        if info is None:
            print("PageID not found.")
        else:
            print("\t".join(info.keys()))
            print("\t".join(map(str,info.values())))

    def search_n(self,n,delta_n):
        print("*Search n =",n,"delta_n =",delta_n)
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        interval = [n-delta_n,n+delta_n]
        c.execute('''select r.pageid,shelf,book,page,r.wave,r.refindex
                    from refractiveindex r join pages p on r.pageid = p.pageid
                    where refindex between ? and ? ''',interval)
        results = c.fetchall()
        if len(results)==0:
            print("No results found.")
        else:
            print(len(results),"results found.")
            print("pageid|shelf|book|page|wavelength|n")
            for r in results:
                print(r)
        conn.close()

    def search_k(self,k,delta_k):
        print("*Search k =",k,"delta_k =",delta_k)
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        interval = [k-delta_k,k+delta_k]
        c.execute('''select e.pageid,shelf,book,page,e.wave,e.coeff
                    from extcoeff e join pages p on e.pageid = p.pageid
                    where coeff between ? and ?''',interval)
        results = c.fetchall()
        if len(results)==0:
            print("No results found.")
        else:
            print(len(results),"results found.")
            print("pageid|shelf|book|page|wavelength|k")
            for r in results:
                print(r)
        conn.close()

    def search_nk(self,n,delta_n,k,delta_k):
        print("*Search n =",n,"delta_n =",delta_n,"k =",k,"delta_k =",delta_k)
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        interval = [n-delta_n,n+delta_n,k-delta_k,k+delta_k]
        c.execute('''select r.pageid, shelf, book, page, r.wave, r.refindex, e.coeff
                    from refractiveindex r join extcoeff e on r.pageid = e.pageid and r.wave = e.wave
                    join pages p on r.pageid = p.pageid
                    where refindex between ? and ? and coeff between ? and ?''',interval)
        results = c.fetchall()
        if len(results)==0:
            print("No results found.")
        else:
            print(len(results),"results found.")
            print("pageid|shelf|book|page|wavelength|n|k")
            for r in results:
                print(r)
        conn.close()

    def get_material(self, pageid):
        pagedata = self._get_page_info(pageid)
        if pagedata is None:
            print("PageID not found.")
            return None
        else:
            conn = sqlite3.connect(self.db_path)
            c = conn.cursor()
            wavelengths_r = None
            wavelengths_e = None
            refractive = None
            extinction = None
            if pagedata['hasrefractive'] == 1:
                c.execute('''select wave,refindex
                            from refractiveindex
                            where pageid = ?
                            order by wave asc''', [pageid])
                results = c.fetchall()
                wavelengths_r = [r[0] for r in results]
                refractive = [r[1] for r in results]
            if pagedata['hasextinction'] == 1:
                c.execute('''select wave,coeff
                            from extcoeff
                            where pageid = ?
                            order by wave asc''', [pageid])
                results = c.fetchall()
                wavelengths_e = [r[0] for r in results]
                extinction = [r[1] for r in results]
            conn.close()
            print("Material",pagedata['filepath'],"loaded.")
            return DBMaterial.FromLists(pagedata,wavelengths_r=wavelengths_r,refractive=refractive,
                                      wavelengths_e=wavelengths_e,extinction=extinction)

    def get_material_n_numpy(self,pageid):
        mat = self.get_material(pageid)
        if mat is None:
            return None
        n = mat.get_complete_refractive()
        if n is None:
            print("Material has no refractive data.")
            return None
        return np.array(n)

    def get_material_k_numpy(self,pageid):
        mat = self.get_material(pageid)
        if mat is None:
            return None
        k = mat.get_complete_extinction()
        if k is None:
            print("Material has no extinction data.")
            return None
        return np.array(k)

    def get_material_csv(self, pageid, output="", folder=""):
        mat = self.get_material(pageid)
        if mat is None:
            print("PageID not found.")
            return None
        matInfo = mat.get_page_info()
        #print(matInfo)
        if output=="":
            output = ",".join([str(matInfo['pageid']),matInfo['shelf'],matInfo['book'],matInfo['page']])+".csv"
        if folder != "":
            output = folder+os.sep+output
        mat.to_csv(output)

    def get_material_txt(self, pageid, output="", folder=""):
        mat = self.get_material(pageid)
        if mat is None:
            print("PageID not found.")
            return None
        matInfo = mat.get_page_info()
        #print(matInfo)
        if output=="":
            output = "_".join([str(matInfo['pageid']),matInfo['shelf'],matInfo['book'],matInfo['page']])+".txt"
        if folder != "":
            output = folder+os.sep+output

        if '.txt' not in output:
            output = output+'.txt'

        mat.to_txt(output)

    def get_material_csv_all(self, outputfolder):
        allids = self._get_all_pageids()
        for id in allids:
            print("Processing",id)
            self.get_material_csv(pageid=id, output="", folder=outputfolder)

    def _get_pages_columns(self):
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        c.execute('PRAGMA table_info(pages);')
        results = c.fetchall()
        names = [r[1] for r in results]
        conn.close()
        return names

    def _get_page_info(self,pageid):
        columns = self._get_pages_columns()
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        c.execute('SELECT * FROM pages WHERE pageid = ?',[pageid])
        results = c.fetchall()
        if len(results) == 0:
            conn.close()
            return None
        else:
            row = results[0]
            data = OrderedDict.fromkeys(columns)
            for idx,c in enumerate(columns):
                data[c] = row[idx]
            #data = {columns[i]:row[i] for i in range(len(columns))}
            conn.close()
            return data

    def _get_all_pageids(self):
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        c.execute('SELECT pageid FROM pages')
        results = c.fetchall()
        if len(results) == 0:
            conn.close()
            return None
        else:
            pageids = [row[0] for row in results]
            return pageids

    @staticmethod
    def DownloadRIIzip(outputfolder="", riiurl=_riiurl):
        import requests, zipfile, io, tempfile

        with tempfile.TemporaryDirectory() as fp:
            print("Making request to",riiurl)
            r = requests.get(riiurl)
            if r.ok:
                print("Downloaded and extracting...")
                z = zipfile.ZipFile(io.BytesIO(r.content))
                z.extractall(path=outputfolder)
                print("Wrote",outputfolder+"/database","from",riiurl)
                #The destination+database is the result.
                result = True
            else:
                print("There was a problem with the request.")
                result = False
        return result

def extract_entry_list(db_path):
    entries = []
    referencePath = os.path.normpath(db_path)
    idx = 0
    with open(os.path.join(referencePath, os.path.normpath("library.yml")), "r") as f:
        catalog = yaml.safe_load(f)
    for sh in catalog:
        shelf = Shelf(sh['SHELF'], sh['name'])
        for b in sh['content']:
            if 'DIVIDER' not in b:
                book = Book(b['BOOK'],b['name'])
                for p in b['content']:
                    if 'DIVIDER' not in p:
                        page = Page(p['PAGE'],
                                    p['name'],
                                    os.path.join(os.path.join(referencePath, 'data'), os.path.normpath(p['data'])))

                        entries.append(Entry(str(idx),shelf,book,page))
                        idx+=1
    return entries

def print_pretty_entry_list(entries):
    for e in entries:
        print(",".join([e.id,e.shelf.shelf,e.book.book,e.page.page]))

def pretty_entry(entry):
    e = entry
    return ",".join([e.id,e.shelf.shelf,e.book.book,e.page.page])

def create_sqlite_database(refractiveindex_db_path,new_sqlite_db,interpolation_points=100):
    conn = sqlite3.connect(new_sqlite_db)
    c = conn.cursor()
    c.execute('''DROP TABLE IF EXISTS pages;''')
    c.execute('''DROP TABLE IF EXISTS refractiveindex;''')
    c.execute('''DROP TABLE IF EXISTS extcoeff;''')
    c.execute('''CREATE TABLE pages
    (pageid int, shelf text COLLATE NOCASE, book text COLLATE NOCASE, page text COLLATE NOCASE,
    filepath text COLLATE NOCASE,
    hasrefractive integer, hasextinction integer,
    rangeMin real, rangeMax real, points int)''')
    c.execute('''CREATE TABLE refractiveindex (pageid int, wave real, refindex real)''')
    c.execute('''CREATE TABLE extcoeff (pageid int, wave real, coeff real)''')
    conn.commit()
    conn.close()
    _populate_sqlite_database(refractiveindex_db_path,new_sqlite_db,interpolation_points=interpolation_points)

def _populate_sqlite_database(refractiveindex_db_path,new_sqlite_db,interpolation_points=100):
    entries = extract_entry_list(refractiveindex_db_path)
    conn = sqlite3.connect(new_sqlite_db)
    c = conn.cursor()
    for e in entries:
        try:
            mat = DBMaterial(filename=e.page.path,interpolation_points=interpolation_points)
            hasrefractive=0
            hasextinction=0
            if mat.has_refractive():
                refr = mat.get_complete_refractive()
                hasrefractive = 1
                values = [[e.id,r[0],r[1]] for r in refr]
                c.executemany('INSERT INTO refractiveindex VALUES (?,?,?)', values)
            if mat.has_extinction():
                ext = mat.get_complete_extinction()
                hasextinction = 1
                values = [[e.id,ex[0],ex[1]] for ex in ext]
                c.executemany('INSERT INTO extcoeff VALUES (?,?,?)', values)
            c.execute("INSERT INTO pages VALUES (?,?,?,?,?,?,?,?,?,?)",
                      [e.id,
                       e.shelf.shelf,
                       e.book.book,
                       e.page.page,
                       os.sep.join(e.page.path.split(os.sep)[-3:]),
                       hasrefractive,
                       hasextinction,
                       mat.rangeMin,
                       mat.rangeMax,
                       mat.points])
        except Exception as error:
            print("LOG:",pretty_entry(e),":",error)
    conn.commit()
    conn.close()
    print("***Wrote SQLite DB on ",new_sqlite_db)


def pipeline_test():
    #Database.DownloadRIIzip()
    db = Database("../refractive.db")

    #db.check_url_version()
    #db.create_database_from_folder(yml_database_path="database", interpolation_points=200)
    #db.create_database_from_url(interpolation_points=200)
    #db.create_database_from_url(riiurl="http://refractiveindex.info/download/database/rii-database-2015-07-05.zip")

    db.search_pages()
    #db.search_pages("otanicar")
    #db.search_pages("au",exact=True)
    #Id 327 for Formula2 test

    #print(db._get_page_info(1542))
    #db.search_id(1542)

    #mat = db.get_material(1542) #Only extinction
    #mat = db.get_material(372) #Both (formula)
    #mat = db.get_material(1) #Both (tabulated)
    #print("HasRefractive?",mat.has_refractive())
    #print("HasExtinction?",mat.has_extinction())
    #print(mat.get_complete_refractive())
    #print(mat.get_complete_extinction())
    #print(mat.get_page_info())
    #mat.to_csv(output="mat1.csv")

    #db.get_material_csv(1542,output="",folder="all")
    #db.get_material_csv_all(outputfolder="all")

    #db.search_n(n=0.3,delta_n=.001)
    #db.search_k(k=0.3,delta_k=.001)
    #db.search_nk(n=0.3, delta_n=0.1,k=0.3,delta_k=0.1)

    #print(db.search_custom('select * from pages where shelf="main" and book="Ag" and page LIKE "%k%"'))
    #print(db.search_custom('select wave,coeff from extcoeff where pageid = 1 and wave between 0.3 and 0.4'))
    #print(db.search_custom('''select p.filepath, r.wave,refindex,coeff
    #                from refractiveindex r inner join extcoeff e on r.pageid = e.pageid and r.wave = e.wave
    #                inner join pages p on r.pageid = p.pageid
    #                where r.wave = .301'''))

if __name__ == '__main__':
    pipeline_test()
